# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Define types for `LeafHandler`."""

import dataclasses
from typing import Any, Awaitable, Generic, Protocol, Sequence, Tuple, Type, TypeVar

from orbax.checkpoint._src.serialization import serialization as serialization_v0
from orbax.checkpoint._src.tree import utils as tree_utils
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types
import tensorstore as ts


Leaf = TypeVar('Leaf')
AbstractLeaf = TypeVar('AbstractLeaf')


@dataclasses.dataclass
class SerializationParam(Generic[Leaf]):
  keypath: tree_types.PyTreeKeyPath
  value: Leaf

  @property
  def name(self) -> str:
    return tree_utils.param_name_from_keypath(self.keypath)


@dataclasses.dataclass
class SerializationContext:
  parent_dir: path_types.PathAwaitingCreation
  ts_context: ts.Context | None = None
  byte_limiter: serialization_v0.LimitInFlightBytes | None = None


@dataclasses.dataclass
class DeserializationParam(Generic[AbstractLeaf]):
  keypath: tree_types.PyTreeKeyPath
  value: AbstractLeaf | None = None

  @property
  def name(self) -> str:
    return tree_utils.param_name_from_keypath(self.keypath)


@dataclasses.dataclass
class DeserializationContext:
  parent_dir: path_types.Path
  ocdbt_checkpoint: bool
  zarr3_checkpoint: bool
  ts_context: ts.Context | None = None
  byte_limiter: serialization_v0.LimitInFlightBytes | None = None


class LeafHandler(Protocol[Leaf, AbstractLeaf]):
  """Interface for reading and writing a PyTree leaf."""

  async def serialize(
      self,
      params: Sequence[SerializationParam[Leaf]],
      serialization_context: SerializationContext,
  ) -> Awaitable[None]:
    """Writes the specified leaves of a checkpointable to a storage location.

    This method is responsible for copying the leaf data from a remote device in
    a synchronous fashion (if applicable).  The whole operation will be started
    in background as soon as the awaitable is returned.  The returned awaitable
    can be awaited to confirm the completion of the final commit operation to a
    storage location.

    The function can be used in a multihost setting, but should not implement
    extra logic to ensure atomicity.

    Note: For function authors, within this function, make sure the destination
    directory is fully created by waiting for `await
    serialization_context.parent_dir.await_creation()` before performing any
    storage write operations.

    Args:
      params: a sequence of SerializationParam per leaf.
      serialization_context: SerializationContext for the leaf handler.

    Returns:
      A awaitable which can be awaited to complete the save operation.
    """
    ...

  async def deserialize(
      self,
      params: Sequence[DeserializationParam[AbstractLeaf]],
      deserialization_context: DeserializationContext,
  ) -> Awaitable[Sequence[Leaf]]:
    """Returns sequence of leaves from a stored checkpointable location.

    This method initiates the background copying of leaf data from a storage
    location to remote devices. The returned awaitable allows you to await it to
    confirm the completion of this data transfer.

    Args:
      params: sequence of DeserializationParam per leaf.
      deserialization_context: DeserializationContext for the leaf handler.

    Returns:
      A awaitable which can be awaited to complete the load operation and obtain
      a squence of leaves.
    """
    ...

  async def metadata(
      self,
      params: Sequence[DeserializationParam[None]],
      deserialization_context: DeserializationContext,
  ) -> Sequence[AbstractLeaf]:
    """Returns a squence of AbstractLeaf from a stored checkpointable location.

    Args:
      params: sequence of DeserializationParam[None] per leaf, the keypath is
        essential to look up the leaf metadata.
      deserialization_context: DeserializationContext for the leaf handler.

    Returns:
      Sequence of AbstractLeaf for each provided DeserializationParam.
    """
    ...


LeafHandlerRegistryItem = Tuple[
    Type[Leaf], Type[AbstractLeaf], LeafHandler[Leaf, AbstractLeaf]
]


class LeafHandlerRegistry(Protocol):
  """A Protocol for a LeafHandlerRegistry.

  This protocol defines the interface for a leaf handler registry. It acts as a
  lookup service, associating specific Leaf or AbstractLeaf types with their
  corresponding leaf handlers. It can be accessed through the module function
  get/set/is_handable/is_abstract_handlable.
  """

  def get(self, ty: Type[Leaf]) -> LeafHandler[Leaf, Any]:
    """Returns the handler registered for a given Leaf type, if available."""
    ...

  def get_abstract(
      self, abstract_ty: Type[AbstractLeaf]
  ) -> LeafHandler[Any, AbstractLeaf]:
    """Returns the handler registered for a given abstract type, if available."""
    ...

  def get_all(
      self,
  ) -> Sequence[LeafHandlerRegistryItem]:
    """Returns all registered handlers. Useful to examine what is registered.

    Returns:
      A sequence of tuples containing the type, abstract type, and handler for
      corresponding registered handler.
    """
    ...

  def add(
      self,
      ty: Type[Leaf],
      abstract_ty: Type[AbstractLeaf],
      handler: LeafHandler[Leaf, AbstractLeaf],
      override: bool = False,
  ):
    """Adds a handler for a given type and abstract type pair.

    If there is already a registered handler for the a given concrete type, its
    coressponding abstract_ty and handler will be overridden. If the abstract
    type has already associated with another concrete type, an error will be
    raised.


    Args:
      ty: The type to register the handler for.
      abstract_ty: The abstract type to register the handler for.
      handler: The handler to register.
      override: Whether to override the handler if it already exists.
    """
    ...

  def is_handleable(self, ty: Type[Any]) -> bool:
    """Returns True if the type is handleable by any registered handler."""
    ...

  def is_abstract_handleable(self, abstract_ty: Type[Any]) -> bool:
    """Returns True if the abstract type is handlable by any registered handler."""
    ...



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""JsonCheckpointHandler class.

Implementation of CheckpointHandler interface.
"""

from __future__ import annotations

import dataclasses
import json
from typing import Any, List, Mapping, Optional

from etils import epath
from orbax.checkpoint import checkpoint_args
from orbax.checkpoint import options as options_lib
from orbax.checkpoint import utils
from orbax.checkpoint._src import asyncio_utils
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.handlers import async_checkpoint_handler

CheckpointArgs = checkpoint_args.CheckpointArgs
register_with_handler = checkpoint_args.register_with_handler


class JsonCheckpointHandler(async_checkpoint_handler.AsyncCheckpointHandler):
  """Saves nested dictionary using json."""

  def __init__(
      self,
      filename: Optional[str] = None,
      *,
      multiprocessing_options: options_lib.MultiprocessingOptions = options_lib.MultiprocessingOptions(),
  ):
    """Initializes JsonCheckpointHandler.

    Args:
      filename: optional file name given to the written file; defaults to
        'metadata'
      multiprocessing_options: See orbax.checkpoint.options.
    """
    self._filename = filename or 'metadata'
    self._primary_host = multiprocessing_options.primary_host

  async def _save_fn(self, x, directory):
    if utils.is_primary_host(self._primary_host):
      path = directory / self._filename
      path.write_text(json.dumps(x))

  async def async_save(
      self,
      directory: epath.Path,
      item: Optional[Mapping[str, Any]] = None,
      args: Optional[JsonSaveArgs] = None,
  ) -> Optional[List[future.Future]]:
    """Saves the given item.

    Args:
      directory: save location directory.
      item: Deprecated, use `args` instead.
      args: JsonSaveArgs (see below).

    Returns:
      A list of commit futures.
    """
    if isinstance(item, CheckpointArgs):
      raise ValueError(
          'Make sure to specify kwarg name `args=` when providing'
          ' `JsonSaveArgs`.'
      )
    if args is not None:
      item = args.item
    return [
        future.CommitFutureAwaitingContractedSignals(
            self._save_fn(item, directory), name='json_ch_save'
        )
    ]

  def save(
      self,
      directory: epath.Path,
      item: Optional[Mapping[str, Any]] = None,
      args: Optional[JsonSaveArgs] = None,
  ):
    async def async_save(directory, item, args):
      commit_futures = await self.async_save(directory, item, args)
      if commit_futures:
        for f in commit_futures:
          f.result()

    asyncio_utils.run_sync(async_save(directory, item, args))

  def restore(
      self,
      directory: epath.Path,
      item: Optional[Mapping[str, Any]] = None,
      args: Optional[JsonRestoreArgs] = None,
  ) -> Mapping[str, Any]:
    """Restores json mapping from directory.

    `item` is unused.

    Args:
      directory: restore location directory.
      item: unused
      args: unused

    Returns:
      JSON dict.

    Raises:
      FileNotFoundError: if the file does not exist.
    """
    del item
    del args
    path = directory / self._filename
    if not path.exists():
      raise FileNotFoundError(f'File {path} not found.')
    return json.loads(path.read_text())


@register_with_handler(JsonCheckpointHandler, for_save=True)
@dataclasses.dataclass
class JsonSaveArgs(CheckpointArgs):
  """Parameters for saving to json.

  Attributes:
    item (required): a nested dictionary.
  """

  item: Mapping[str, Any]


@register_with_handler(JsonCheckpointHandler, for_restore=True)
@dataclasses.dataclass
class JsonRestoreArgs(CheckpointArgs):
  """Json restore args.

  Attributes:
    item: unused, but included for legacy-compatibility reasons.
  """

  item: Optional[bytes] = None



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""`PartsOf[T]` holds a data structure `T` whose leaves may be missing.

`Optional[T]` is equivalent to `T | None` and represents either an entire `T`
or no value. `PartsOf[T]` lies somewhere between those two extremes.

Not to be confused with `jax_tree_utils.Partial`, which does partial function
application.
"""

import collections
import dataclasses
from typing import Any, Callable, Generic, Mapping, TypeVar

from absl import logging
from jax import tree_util as jax_tree_util
from orbax.checkpoint._src.tree import utils as tree_utils

T = TypeVar('T')

# Represents a structure that is PyTree-compatible with a `T` but whose leaves
# may be placeholders.
PyTreeWithPlaceholdersLike = Any | T


# The value used to represent a missing leaf. This specific value is useful
# because:
#   1. JAX does not treat it as an internal node (which it does with `None`);
#   2. There is only one of its type, so `is`-based comparison works.
PLACEHOLDER = ...
Placeholder = type(PLACEHOLDER)


class Error(ValueError):
  pass


class TreeNotCompleteError(Error):
  pass


class UnexpectedPlaceholderError(Error):
  pass


def _check_templates_match(t0: Any, t1: Any) -> None:
  if t0 != t1:
    logging.warning(
        'Mismatched templates: \nCurrent Template - t0= %s \nOther Template -'
        ' t1= %s ',
        t0,
        t1,
    )
    raise Error('Mismatched templates')


@dataclasses.dataclass(init=False)
class PartsOf(Generic[T]):
  """Holds a data structure `T` whose leaves may be missing.

  `T` must be understandable as a JAX PyTree.

  The only operations that may safely be performed on a `PartsOf[T]` are:
  - Element-wise combination with another `PartsOf[T]`;
  - Checked extraction of a complete structure.

  For example:
  ```
  def fetch_all() -> MyStructure:
    some_parts: PartsOf[MyStructure] = fetch_from_one_place()
    other_parts: PartsOf[MyStructure] = fetch_from_another_place()
    return some_parts.overlay(other_parts).full_structure
  ```

  A `PartsOf[T]` can be introduced by constructing it with a template `T` and
  a value that supplies none, some or all of the leaves of T. Every path to a
  leaf in the value must correspond to a path to a leaf in the template.
  (This is not what is usually meant by the term 'subtree' and perhaps the
  term 'recursive subset' would best describe the relationship between the
  value and the template.)

  For example:
  ```
  @dataclasses.dataclass
  class MyStructure:
    x: int
    y: dict[str, int]

  template = MyStructure(x=0, y={'one': 1, 'two': 2})
  value = {'y': {'two': 4}}

  partial_structure = PartsOf(template, value)
  ```

  In the case where the value supplies no leaves at all,
  `PartsOf.empty(template)` is an available shorthand.
  """

  _template: jax_tree_util.PyTreeDef
  _present: dict[tuple[Any, ...], Any]

  def __init__(
      self,
      template: T,
      value: PyTreeWithPlaceholdersLike[T],
  ):
    unused_leaves, self._template = jax_tree_util.tree_flatten(template)
    self._present = {
        k: v
        for k, v in tree_utils.to_flat_dict(value).items()
        if v is not PLACEHOLDER
    }

    template_paths = tree_utils.to_flat_dict(template)
    bogus_paths = {*self._present} - {*template_paths}
    if bogus_paths:
      raise Error(
          'The following paths in input value are not part of the full tree'
          f' structure: {bogus_paths}.\nGot value paths {self._present}\nand'
          f' template paths {template_paths}'
      )

  @classmethod
  def empty(
      cls,
      template: T,
  ) -> 'PartsOf[T]':
    return cls(template, ())

  @classmethod
  def from_structure(
      cls,
      structure: T
  ) -> 'PartsOf[T]':
    """Makes a PartsOf from a structure using it as both template and value.

    Args:
      structure: The structure to use as both template and value. May contain
        placeholders for missing leaves.

    Returns:
      A PartsOf representing the given structure.
    """
    return cls(structure, structure)

  # Alias for backwards compatibility.
  from_full_structure = from_structure

  @classmethod
  def like(
      cls,
      existing: 'PartsOf[Any]',
      value: PyTreeWithPlaceholdersLike[T],
  ) -> 'PartsOf[T]':
    return cls(
        existing._get_template(),  # pylint:disable=protected-access
        value,
    )

  @classmethod
  def _with_present_values(
      cls,
      template: jax_tree_util.PyTreeDef,
      present_values: dict[tuple[Any, ...], Any],
  ):
    result = cls.empty(())
    result._template = template
    result._present = present_values
    return result

  def _get_template(self):
    return self._template.unflatten([PLACEHOLDER] * self._template.num_leaves)

  def is_empty(self) -> bool:
    """Returns True if the PartsOf is empty."""
    return not self._present

  @property
  def structure_hash(self) -> int:
    """Returns a hash of the structure."""
    return hash((self._template, tuple(self._present.keys())))

  @property
  def full_structure(self) -> T:
    """Extracts a full PyTree with no missing leaves, or raises Error."""
    try:
      return tree_utils.from_flat_dict(self._present, self._get_template())
    except KeyError as e:
      logging.warning(
          'Detected mismatch between values and template.\n'
          'Value keys: %s\n'
          'Template: %s',
          self._present.keys(),
          self._get_template(),
      )
      raise TreeNotCompleteError(f'Tree is not complete: {e}') from e

  @property
  def unsafe_structure(self) -> PyTreeWithPlaceholdersLike[T]:
    """Extracts a `T` that may have missing leaves. Avoid this if possible."""
    leaves = {
        **tree_utils.to_flat_dict(self._get_template()),
        **self._present,
    }
    return tree_utils.from_flat_dict(leaves, self._get_template())

  def overlay(
      self,
      other: 'PartsOf[T]',
      *,
      strict: bool = False,
      verbose: bool = False,
  ) -> 'PartsOf[T]':
    """Overlay another partial tree on this one.

    Args:
      other: The partial tree to overlay. Must have the same template as this
        one.
      strict: If True, the overlay operation will fail if sets of values are
        not disjoint in this tree and `other`.
      verbose: Whether to log the keys that are being overwritten from `other`.

    Returns:
      A new `PartsOf[T]` where a set of values has been updated with values from
      `other`.
      If a value for some tree path is present in both trees:
        - if `strict` is False: the value from `other` is used (overwrites the
          value from this tree).
        - if `strict` is True: an error is raised.

      Example:
        ```
        t = {'a': 1, 'b': 2, 'c': ...}
        t_other = {'a': ..., 'b': 13, 'c': 4}

        t.overlay(t_other)  # Succeeds, returns {'a': 1, 'b': 13, 'c': 4}.
        t.overlay(t_other, strict=True)  # Raises (can't overwrite 'b').
        ```
    """
    _check_templates_match(self._template, other._template)  # pylint:disable=protected-access
    common_keys = self._present.keys() & other._present.keys()  # pylint:disable=protected-access
    if strict:
      if common_keys:
        raise ValueError(
            f'Strict overlay failed: keys {common_keys} are present in both'
            ' PartsOf structures.'
        )
    elif verbose:
      if common_keys:
        logging.info('Overwriting %d keys:', len(common_keys))
        for key in sorted(list(common_keys)):
          logging.info('  %s', key)
      else:
        logging.info('No keys to overwrite.')
    new_leaves = self._present.copy()
    new_leaves.update(other._present)  # pylint:disable=protected-access
    return PartsOf._with_present_values(self._template, new_leaves)

  def intersect(self, other: 'PartsOf[T]') -> 'PartsOf[T]':
    """Removes all leaves that are not present in `other`."""
    _check_templates_match(self._template, other._template)  # pylint:disable=protected-access
    new_leaves = {k: v for k, v in self._present.items() if k in other._present}  # pylint:disable=protected-access
    return PartsOf._with_present_values(self._template, new_leaves)

  def __sub__(self, other: 'PartsOf[T]') -> 'PartsOf[T]':
    """Removes all leaves that are present in `other`."""
    _check_templates_match(self._template, other._template)
    new_leaves = {
        k: v for k, v in self._present.items() if k not in other._present
    }
    return PartsOf._with_present_values(self._template, new_leaves)

  def tree_flatten(self) -> tuple[tuple[T, ...], jax_tree_util.PyTreeDef]:
    """JAX pytree flatten implementation for PartsOf."""
    return (self._present,), self._template

  @classmethod
  def tree_unflatten(
      cls,
      aux_data: jax_tree_util.PyTreeDef,
      children: tuple[dict[tuple[Any, ...], Any]],  # pylint:disable=g-one-element-tuple
  ) -> 'PartsOf[T]':
    """JAX pytree unflatten implementation for PartsOf."""
    template = aux_data
    (present,) = children
    for p, v in present.items():
      if v is PLACEHOLDER:
        raise UnexpectedPlaceholderError(
            'PartsOf tree unflattening discovered a placeholder value at path'
            f' {p}, which is not supported. If you need to map over a PartsOf'
            ' tree *returning placeholder values*, please use'
            ' `parts_of.map_leaves` or `parts_of.map_leaves_with_path` instead'
            ' of the standard JAX tree mapping functions.'
        )
    return PartsOf._with_present_values(template, present)


jax_tree_util.register_pytree_node_class(PartsOf)


def map_leaves(
    f: Callable[..., Any | Placeholder],
    *xs: PartsOf[T],
) -> PartsOf[T]:
  """Applies a function to all leaves (present or not).

  PartsOf-compatible equivalent of `jax.tree.map`.

  Args:
    f: The function to apply to each leaf of the structure. It must accept as
        many arguments as given here, any number of which may be `PLACEHOLDER`.
        It must return an appropriate non-`None` leaf value, or `PLACEHOLDER`.
    *xs: One or more partially known structures.

  Returns:
    A partially known structure.
  """
  def _wrapped_f(path, *leaves):
    del path
    return f(*leaves)

  return map_leaves_with_path(_wrapped_f, *xs)


def map_leaves_with_path(
    f: Callable[..., Any | Placeholder],
    *xs: PartsOf[T],
) -> PartsOf[T]:
  """Applies a function to all leaves (present or not), with value path.

  PartsOf-compatible equivalent of `jax_tree_util.tree_map_with_path`.

  Args:
    f: The function to apply to each leaf of the structure. It must accept tree
      path + as many arguments as *xs given here, any number of which may be
      `PLACEHOLDER`. It must return an appropriate non-`None` leaf value, or
      `PLACEHOLDER`.
    *xs: One or more partially known structures.

  Returns:
    A partially known structure.
  """
  x0, *xs = xs
  for x in xs:
    _check_templates_match(x0._template, x._template)  # pylint:disable=protected-access
  def f_(path, _):
    path = tree_utils.tuple_path_from_keypath(path)
    leaves = [x._present.get(path, PLACEHOLDER) for x in (x0, *xs)]  # pylint:disable=protected-access
    result = f(path, *leaves)
    if result is None:
      raise Error(f'At {path}, {f} returned None.')
    if not jax_tree_util.all_leaves([result]):
      raise Error(f'At {path}, {f} returned non-leaf value {result}.')
    return result
  t0 = x0._get_template()  # pylint:disable=protected-access
  return PartsOf(t0, jax_tree_util.tree_map_with_path(f_, t0))


def filter_values(
    existing: 'PartsOf[T]',
    value_keys: set[tuple[Any, ...]],
) -> 'PartsOf[T]':
  """Makes a new PartsOf from existing with only the values with given keys."""
  # pylint:disable=protected-access
  result = PartsOf.empty(existing._get_template())
  result._present = {
      k: v for k, v in existing._present.items() if k in value_keys
  }
  # pylint:enable=protected-access
  return result


def value_key_from_path(path: tuple[Any, ...]) -> tuple[Any, ...]:
  """Converts a PartsOf JAX pytree path into a key for the corresponding value.

  Args:
    path: a JAX tree path into a PartsOf object (i.e. obtained via one of
      `jax_tree_util.*` methods on a PartsOf pytree).

  Returns:
    A key for the corresponding value in PartsOf template's flattened dict
    (i.e. `_present`).
  """
  assert len(path) == 2, (
      f'Too many elements in a PartsOf tree path: {path}, expected 2.'
  )
  assert isinstance(
      path[1], jax_tree_util.DictKey
  ), f'Expected DictKey, found: {path[1]}'
  return path[1].key


C = TypeVar('C')


def partition(
    classifier: Callable[[T], C],
    structure: PartsOf[T],
) -> Mapping[C, PartsOf[T]]:
  """Partitions a PartsOf[T] based on a leaf classifier function.

  Args:
    classifier: A function that assigns a class label to each leaf of the
      structure.
    structure: The structure to partition.

  Returns:
    A mapping from classification result to a PartsOf[T] containing only the
    values that were classified as such.

  Example:
    ```
    template = MyDataclass(a=(X, X), b={'c': X, 'd': X})
    t = PartsOf(template, MyDataclass(a=(1, 2), b={'c': 4, 'd': 3}))
    # Split the structure into one with even and another with odd values.
    even_odd = partition(lambda x: x % 2, t)

    assert even_odd[0].unsafe_structure == (
        MyDataclass(a=(..., 2), b={'c': 4, 'd': ...}
    )
    assert even_odd[1].unsafe_structure == (
        MyDataclass(a=(1, ...), b={'c': ...,' d': 3}
    )
    ```
  """
  value_paths_by_class = collections.defaultdict(set)
  for k, v in structure._present.items():  # pylint:disable=protected-access
    c = classifier(v)
    value_paths_by_class[c].add(k)
  return {
      c: filter_values(structure, value_paths)
      for c, value_paths in value_paths_by_class.items()
  }



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Emergency checkpointing utils for multihost / multislice."""

from typing import List, Optional

from absl import logging
import jax
import numpy as np
from orbax.checkpoint._src.multihost import multihost


def _int_list_flip_index_and_value(int_list: List[int]):
  """Reverses indices and values of a list of integers.

  Ex: [ 3, 4, 0, 1, 2 ] -> [ 2, 3, 4, 0, 1 ]
  Old index 0, value: 3 -> New index: 3, value: 0

  Args:
    int_list: List of integers.

  Returns:
    List of integers with reversed indices and values.
  """
  result = [None for _ in range(len(int_list))]
  for index, value in enumerate(int_list):
    result[value] = index
  assert None not in result
  return result


def _get_runtime_id_across_restarts(
    previous_runtime_to_dist_id: Optional[List[int]],
) -> List[Optional[int]]:
  """Get runtime id changes across restarts.

  Args:
    previous_runtime_to_dist_id: mapping from runtime process index to
      distributed process index of the previous incarnation.

  Returns:
    Integer list which maps previous runtime id (index) to current runtime id
    (array element).
  Raises:
    ValueError:
  """
  current_dist_to_runtime_id = _int_list_flip_index_and_value(
      runtime_to_distributed_ids()
  )
  previous_dist_to_runtime_id = _int_list_flip_index_and_value(
      previous_runtime_to_dist_id
  )

  result = [None for _ in range(jax.process_count())]
  # Previous runtime id (index) to current runtime id (value).
  for i in range(jax.process_count()):
    result[previous_dist_to_runtime_id[i]] = current_dist_to_runtime_id[i]
  assert None not in result
  return result


def process_index_from_device_id(device_id: int) -> int:
  """Get process index from device id."""
  if jax.devices()[0].platform == 'gpu':
    return device_id // jax.local_device_count()
  elif jax.devices()[0].platform == 'tpu':
    if hasattr(jax.devices()[0], 'slice_index'):
      # Note that it is possible for single slice TPU devices to have  a slice
      # index.
      num_slices = max([d.slice_index for d in jax.devices()]) + 1
      # Multi-slice TPU workload.
      if num_slices > 1:
        num_processes_per_slice = jax.process_count() // num_slices
        # This is based on how Megascale device ids are assigned.
        # See platforms/xla/megascale/runtime/common/multi_slice_topology.h.
        slice_id = device_id // 100000 - 1
        local_process_id = device_id % 100000 // jax.local_device_count()
        return slice_id * num_processes_per_slice + local_process_id
    # Single slice TPU workload.
    return device_id // jax.local_device_count()
  # CPU workload.
  else:
    # This is based on how CPU device ids are assigned.
    # See tensorflow/compiler/xla/pjrt/cpu/cpu_topology.h.
    return device_id // (1 << 17) // jax.local_device_count()


def consistent_restore_mesh(
    devices: List[jax.Device],
    user_mesh: jax.sharding.Mesh,
    previous_flattened_mesh_device_ids: List[int],
    previous_distributed_to_device_ids: List[List[int]],
    current_distributed_to_device_ids: List[List[int]],
):
  """Create a mesh that is consistent with the previous incarnation.

    1. We can think of mesh as being backed by a list of devices.
    2. Default mesh follows the default device id order [0, ..., n-1]. Or
       the user may permute it according to their needs.
    3. After restart, the user will construct the same software mesh as (2).
    4. But a given hardware device may change its id because of scheduler
       or runtime quirks.
    5. Goal: construct the mesh with the same hardware device order as
       before restart, that may not follow the current software ids.
    6. Thus, we shuffle the device order within the mesh by checking how
       each device's software ids changed across restarts.

  This is to restore the same global array values from local hardware devices
  even if software process and device ids have changed across restarts.

  Args:
    devices: List of Jax devices (usually `jax.devices()`).
    user_mesh: The user mesh.
    previous_flattened_mesh_device_ids: The flattened device ids of the mesh.
    previous_distributed_to_device_ids: The distributed id to range of device
      ids mapping of the previous incarnation.
    current_distributed_to_device_ids: The distributed id to range of device ids
      mapping of the current incarnation.

  Returns:
    The new mesh devices that should be used to create the mesh.
  """
  # Map how device ids changed across restarts.
  device_id_across_restarts = {}
  for i in range(len(previous_distributed_to_device_ids)):
    for j in range(len(previous_distributed_to_device_ids[i])):
      previous_id = previous_distributed_to_device_ids[i][j]
      current_id = current_distributed_to_device_ids[i][j]
      device_id_across_restarts[previous_id] = current_id
  logging.debug(
      'device_id_across_restarts (key: previous_id, value: current_id): %s',
      device_id_across_restarts,
  )
  # Key devices by id.
  jax_devices_by_id = {d.id: d for d in devices}

  new_flattened_mesh_devices = [
      # Convert old ids to current ids that correspond to the same physical
      # hardware.
      jax_devices_by_id[device_id_across_restarts[id]]
      for id in previous_flattened_mesh_device_ids
  ]
  new_mesh_devices = np.array(new_flattened_mesh_devices).reshape(
      user_mesh.devices.shape
  )
  return jax.sharding.Mesh(new_mesh_devices, user_mesh.axis_names)


def runtime_to_distributed_ids() -> List[int]:
  """Returns the runtime to distributed process id mapping."""
  # TODO(b/325293150): Deprecate this after jaxlib contains the fix.
  result = multihost.runtime_to_distributed_ids()
  runtime_and_distributed_ids_are_the_same = all([
      result[i] == i for i in range(len(result))
  ])

  # JAX may choose to overwrite the device process index with the distributed
  # process index. In that case, we have to use the device id to infer the real
  # device process index. This is a hack, the intent is to remove it once this
  # workaround is no longer needed.
  if runtime_and_distributed_ids_are_the_same:
    result = [-1 for _ in range(jax.process_count())]
    devices = jax.devices()
    for i in range(0, jax.device_count(), jax.local_device_count()):
      result[process_index_from_device_id(devices[i].id)] = devices[
          i
      ].process_index
    assert -1 not in result
  return result



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A class providing functionalities for managing a series of checkpoints."""

from __future__ import annotations

import concurrent
import dataclasses
import datetime
import threading
import time
import typing
from typing import Any, Callable, Container, Iterable, List, Mapping, Optional, Sequence, Tuple, Type, Union, overload

from absl import logging
from etils import epath
from etils import epy
import jax
from jax.experimental.array_serialization import serialization as jax_serialization
from orbax.checkpoint import abstract_checkpoint_manager
from orbax.checkpoint import args as args_lib
from orbax.checkpoint import checkpoint_args
from orbax.checkpoint import options as options_lib
from orbax.checkpoint import utils
from orbax.checkpoint._src import threading as threading_lib
from orbax.checkpoint._src.checkpoint_managers import policy_checkpoint_info
from orbax.checkpoint._src.checkpoint_managers import preservation_policy as preservation_policy_lib
from orbax.checkpoint._src.checkpoint_managers import save_decision_policy as save_decision_policy_lib
from orbax.checkpoint._src.checkpointers import abstract_checkpointer
from orbax.checkpoint._src.checkpointers import async_checkpointer
from orbax.checkpoint._src.checkpointers import checkpointer as checkpointer_lib
from orbax.checkpoint._src.futures import synchronization
from orbax.checkpoint._src.handlers import checkpoint_handler
from orbax.checkpoint._src.handlers import composite_checkpoint_handler
from orbax.checkpoint._src.handlers import handler_registration
from orbax.checkpoint._src.handlers import json_checkpoint_handler
from orbax.checkpoint._src.handlers import proto_checkpoint_handler
from orbax.checkpoint._src.logging import abstract_logger
from orbax.checkpoint._src.logging import standard_logger
from orbax.checkpoint._src.logging import step_statistics
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import checkpoint_info
from orbax.checkpoint._src.metadata import root_metadata_serialization
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.path import atomicity_types
from orbax.checkpoint._src.path import deleter
from orbax.checkpoint._src.path import step as step_lib
from orbax.checkpoint._src.path import utils as path_utils
from typing_extensions import Self  # for Python version < 3.11



PyTree = Any
CheckpointDirs = Tuple[str, str]
SaveParams = Mapping[str, Any]
RestoreParams = SaveParams
CheckpointInfo = checkpoint_info.CheckpointInfo
AbstractCheckpointer = abstract_checkpointer.AbstractCheckpointer
CheckpointersDict = Mapping[str, AbstractCheckpointer]
AbstractCheckpointManager = (
    abstract_checkpoint_manager.AbstractCheckpointManager
)
StepMetadata = checkpoint.StepMetadata
RootMetadata = checkpoint.RootMetadata
ItemMetadata = checkpoint.CompositeItemMetadata | checkpoint.SingleItemMetadata
AsyncCheckpointer = async_checkpointer.AsyncCheckpointer
Checkpointer = checkpointer_lib.Checkpointer
JsonCheckpointHandler = json_checkpoint_handler.JsonCheckpointHandler
ProtoCheckpointHandler = proto_checkpoint_handler.ProtoCheckpointHandler
CompositeCheckpointHandler = (
    composite_checkpoint_handler.CompositeCheckpointHandler
)
CheckpointHandler = checkpoint_handler.CheckpointHandler
CheckpointArgs = checkpoint_args.CheckpointArgs
CheckpointHandlersDict = Mapping[str, CheckpointHandler]
CheckpointHandlerRegistry = handler_registration.CheckpointHandlerRegistry
serialize_root_metadata = root_metadata_serialization.serialize
deserialize_root_metadata = root_metadata_serialization.deserialize

AsyncOptions = options_lib.AsyncOptions
MultiprocessingOptions = options_lib.MultiprocessingOptions
FileOptions = options_lib.FileOptions

DEFAULT_ITEM_NAME = 'default'
METRIC_ITEM_NAME = 'metrics'
METADATA_ITEM_NAME = 'metadata'
RESERVED_ITEM_NAMES = []

_INIT_TIME = datetime.datetime.now(tz=datetime.timezone.utc)


def _metrics_file_exists(metrics_item_path: epath.Path) -> bool:
  """True if item directory AND actual file both exist."""
  return (
      metrics_item_path.exists()
      and (metrics_item_path / METRIC_ITEM_NAME).exists()
  )


class StepAlreadyExistsError(ValueError):
  """Raised when a step is already present for a save request."""

  pass


class _FinalizeThread(threading.Thread):
  """Thread wrapper that raises an exception if encountered."""

  exception = None

  def __init__(
      self,
      step: int,
      target: Callable[..., object],
      name: str,
      args=(),
      kwargs=None,
      *,
      daemon=None,
  ):
    super().__init__(
        target=target,
        name=name,
        args=args,
        kwargs=kwargs,
        daemon=daemon,
    )
    self._step = step
    self._thread_local = threading.local()

  def step(self) -> int:
    return self._step

  def run(self):
    try:
      super().run()
    except BaseException as e:  # pylint:disable=broad-exception-caught
      self.exception = e

  def join(self, *args, **kwargs):
    super().join(*args, **kwargs)

    # Keep track of whether the calling thread should raise the exception.
    if not hasattr(self._thread_local, 'should_raise'):
      self._thread_local.should_raise = True
    if self.exception and self._thread_local.should_raise:
      # Each thread should only raise once.
      self._thread_local.should_raise = False
      raise self.exception


@dataclasses.dataclass
class _ShouldSaveFnPolicy(save_decision_policy_lib.SaveDecisionPolicy):
  """A policy that uses a user-provided should_save_fn."""

  should_save_fn: Callable[[int, Optional[int]], bool]

  def should_save(
      self,
      step: policy_checkpoint_info.PolicyCheckpointInfo,
      previous_steps: Sequence[policy_checkpoint_info.PolicyCheckpointInfo],
      *,
      context: save_decision_policy_lib.DecisionContext,
  ) -> bool:
    return self.should_save_fn(
        step.step, previous_steps[-1].step if previous_steps else None
    )


def _get_default_save_decision_policy(
    options: CheckpointManagerOptions,
) -> save_decision_policy_lib.SaveDecisionPolicy:
  """Creates a default policy from CheckpointManagerOptions."""
  save_interval_policies = []
  if options.should_save_fn is not None:
    save_interval_policies.append(_ShouldSaveFnPolicy(options.should_save_fn))
    save_interval_policies.append(
        save_decision_policy_lib.PreemptionCheckpointingPolicy()
    )
  else:
    if options.save_interval_steps is not None:
      save_interval_policies.append(
          save_decision_policy_lib.FixedIntervalPolicy(
              options.save_interval_steps
          )
      )
    if options.save_on_steps is not None:
      save_interval_policies.append(
          save_decision_policy_lib.SpecificStepsPolicy(options.save_on_steps)
      )
    save_interval_policies.append(
        save_decision_policy_lib.PreemptionCheckpointingPolicy()
    )
    save_interval_policies.append(save_decision_policy_lib.InitialSavePolicy())
  return save_decision_policy_lib.AnySavePolicy(save_interval_policies)


def _get_default_preservation_policy(
    options: CheckpointManagerOptions,
) -> preservation_policy_lib.PreservationPolicy:
  """Returns a default preservation policy."""
  preservation_policies = []
  if options.keep_period is not None:
    preservation_policies.append(
        preservation_policy_lib.EveryNSteps(options.keep_period)
    )
  if options.keep_time_interval is not None:
    total_seconds = int(options.keep_time_interval.total_seconds())
    preservation_policies.append(
        preservation_policy_lib.EveryNSeconds(interval_secs=total_seconds)
    )
  if options.best_fn is not None:
    preservation_policies.append(
        preservation_policy_lib.BestN(
            get_metric_fn=options.best_fn,
            reverse=(options.best_mode == 'min'),
            n=options.max_to_keep,
            keep_checkpoints_without_metrics=options.keep_checkpoints_without_metrics,
        )
    )
  else:
    preservation_policies.append(
        preservation_policy_lib.LatestN(n=options.max_to_keep)
    )
  return preservation_policy_lib.AnyPreservationPolicy(preservation_policies)


# TODO(b/268051457) Clean up when no longer depended upon by internal users.
def is_async_checkpointer(checkpointer: AbstractCheckpointer):
  return isinstance(
      checkpointer, async_checkpointer.AsyncCheckpointer
  ) or isinstance(
      checkpointer,
      jax_serialization.GlobalAsyncCheckpointManagerBase,
  )


# TODO(b/309965339) Set todelete_subdir defaults if directory is on CNS.
@dataclasses.dataclass
class CheckpointManagerOptions:
  """Optional arguments for :py:class:`.CheckpointManager`.

  save_interval_steps:
    The interval at which checkpoints should be saved.
    Ensures checkpoints will only be saved every n steps. Defaults to 1.
  max_to_keep:
    deprecated, do not use. use `preservation_policy` instead.
  keep_time_interval:
    deprecated, do not use. use `preservation_policy` instead.
  keep_period:
    deprecated, do not use. use `preservation_policy` instead.
  should_keep_fn:
    deprecated, do not use. use `preservation_policy` instead.
  best_fn:
    If set, maintains checkpoints based on the quality of given
    metrics rather than recency. The function should accept a PyTree of metrics,
    and return a scalar value that can be used to determine the quality score
    of the checkpoint. If `max_to_keep` is also set, then the retained
    checkpoints will be kept based on their quality, as measured by this
    function.
  best_mode:
    One of ['max', 'min']. The best metric is determine on the basis of this
    value.
  keep_checkpoints_without_metrics:
    If False, checkpoints without metrics present
    are eligible for cleanup. Otherwise, they will never be deleted.
  step_prefix:
    If provided, step directories will take the form
    f'{step_prefix}_<step>'. Otherwise, they will simply be an integer <step>.
  step_format_fixed_length:
    If set, formats step with n digits (leading zeros).
    This makes sorting steps easier. Otherwise, step has no leading zeros.
  step_name_format:
    NameFormat to build or find steps under input root directory. If provided,
    `step_prefix`, `step_format_fixed_length` are ignored.
  create:
    If True, creates the top-level directory if it does not already exist.
  cleanup_tmp_directories:
    If True, cleans up any existing temporary directories
    on CheckpointManager creation.
  save_on_steps:
    Optional set of steps at which checkpoints should be saved.
    Useful to save checkpoints on a fixed set of steps that are not multiple of
    `save_interval_steps`.
  single_host_load_and_broadcast:
    If True, calling `all_steps(read=True)` will load on only a single host, and
    will then be broadcast to other hosts. Otherwise, I/O will be performed on
    every host. This can be helpful to reduce QPS to the filesystem if there
    are a large number of hosts.
  todelete_subdir: If set, checkpoints to be deleted will be only renamed into a
    subdirectory with the provided string. Otherwise, they will be directly
    deleted from the file system. Useful if checkpoint deletion is time
    consuming. By default, delete the checkpoint assets. Ignored if file system
    is Google Cloud Storage (directory is prefixed with gs://)
  enable_hns_rmtree: If True, enables additional step of HNS bucket empty folder
    deletion.
  enable_background_delete: If True, old checkpoint deletions will be done in a
    background thread, otherwise, it will be done at the end of each save.  When
    it's enabled, make sure to call CheckpointManager.close() or use context to
    make sure all old steps are deleted before exit.
  read_only: If True, then checkpoints save and delete are skipped. However,
    checkpoints restore works as usual.
  enable_async_checkpointing:
    If True, enables async checkpointing.
  async_options:
    Used to configure properties of async behavior. See above.
  multiprocessing_options: MultiprocessingOptions instance to configure
    multiprocessing behavior.
  should_save_fn:
    Predicate callable to check if given step can be saved. This callable
    accepts step number and optional latest step number as param and returns
    bool. If present then `save_interval_steps` and `save_on_steps` options are
    ignored.
  file_options: Options to configure checkpoint directories and files.
    default=FileOptions().
  save_root_metadata: If True, saves root-level metadata about checkpoints.
    This metadata is not step-specific and is written only once.
  temporary_path_class:
    Optional. The concrete :py:class:`.atomicity_types.TemporaryPath` class to
    be used by the underlying :py:class:`.Checkpointer`.
  save_decision_policy: An object used to determine when a checkpoint should be
    saved. If provided, overrides any other options dealing with this subject,
    including `save_interval_steps`, `save_on_steps`, and `should_save_fn`, and
    is the sole means of determining when a checkpoint should be saved. If not
    provided, these other options are used instead. Prefer to use this option
    over others.
  preservation_policy: An object used to determine which checkpoints to
    preserve. If provided, overrides any other options dealing with this
    subject, including `max_to_keep`, `keep_time_interval`, `keep_period`, and
    `should_keep_fn`, `best_fn`, and is the sole means of determining which
    checkpoints to preserve. If not provided, these other options are used
    instead. Prefer to use this option over others.
  prevent_write_metrics: False by default. If True, metrics will not be written.
  enable_should_save_is_saving_in_progress_check: True by default. If False,
    `should_save_fn` will not check `is_saving_in_progress`, and will assume
    that no save is in progress. This only affects users of
    `ContinuousCheckpointingPolicy` - otherwise the value is ignored.
    This is an interim workaround for b/428061876. Do not use
    without explicit approval.
  """

  save_interval_steps: int = 1
  max_to_keep: Optional[int] = None
  keep_time_interval: Optional[datetime.timedelta] = None
  keep_period: Optional[int] = None
  should_keep_fn: Optional[Callable[[int], bool]] = None
  best_fn: Optional[Callable[[PyTree], float]] = None
  best_mode: str = 'max'
  keep_checkpoints_without_metrics: bool = True
  step_prefix: Optional[str] = None
  step_format_fixed_length: Optional[int] = None
  step_name_format: Optional[step_lib.NameFormat[step_lib.Metadata]] = None
  create: bool = True
  cleanup_tmp_directories: bool = False
  save_on_steps: Optional[Container[int]] = None
  single_host_load_and_broadcast: bool = False
  todelete_subdir: Optional[str] = None
  enable_hns_rmtree: bool = False
  enable_background_delete: bool = False
  read_only: bool = False
  enable_async_checkpointing: bool = True
  async_options: Optional[AsyncOptions] = None
  multiprocessing_options: MultiprocessingOptions = dataclasses.field(
      default_factory=MultiprocessingOptions
  )
  should_save_fn: Optional[Callable[[int, Optional[int]], bool]] = None
  file_options: FileOptions = dataclasses.field(default_factory=FileOptions)
  save_root_metadata: bool = True
  temporary_path_class: Optional[Type[atomicity_types.TemporaryPath]] = None
  save_decision_policy: Optional[
      save_decision_policy_lib.SaveDecisionPolicy
  ] = None
  preservation_policy: Optional[preservation_policy_lib.PreservationPolicy] = (
      None
  )
  prevent_write_metrics: bool = False
  # TODO(b/428061876) Remove this option.
  enable_should_save_is_saving_in_progress_check: bool = True

  def __post_init__(self):
    step_name_format_single_host_load_and_broadcast = (
        hasattr(self.step_name_format, 'single_host_load_and_broadcast')
        and self.step_name_format.single_host_load_and_broadcast
    )
    if self.single_host_load_and_broadcast and self.step_name_format:
      if not step_name_format_single_host_load_and_broadcast:
        raise ValueError(
            '`CheckpointManagerOptions.single_host_load_and_broadcast=True`'
            ' requires `step_name_format.single_host_load_and_broadcast` to be'
            ' set to True.'
        )
    if (
        step_name_format_single_host_load_and_broadcast
        and not self.single_host_load_and_broadcast
    ):
      raise ValueError(
          '`step_name_format.single_host_load_and_broadcast=True` but'
          ' `CheckpointManagerOptions.single_host_load_and_broadcast=False`.'
          ' Please set'
          ' CheckpointManagerOptions.single_host_load_and_broadcast=True.'
      )
    if self.best_mode not in ('min', 'max'):
      msg = (
          "`CheckpointManagerOptions.best_mode` must be one of None, 'min' "
          f"or 'max'. Got {self.best_mode}."
      )
      raise ValueError(msg)
    if (
        self.preservation_policy is None
        and self.max_to_keep is not None
        and self.max_to_keep < 0
    ):
      raise ValueError('Setting of `max_to_keep` must be None or non-negative.')
    if self.read_only and self.save_interval_steps > 0:
      self.save_interval_steps = 0
      logging.warning(
          'CheckpointManagerOptions.read_only=True, setting'
          ' save_interval_steps=0.'
      )
    if self.read_only and self.max_to_keep is not None:
      self.max_to_keep = None
      logging.warning(
          'CheckpointManagerOptions.read_only=True, setting max_to_keep=None.'
      )
    if self.read_only and self.keep_time_interval is not None:
      self.keep_time_interval = None
      logging.warning(
          'CheckpointManagerOptions.read_only=True, setting'
          ' keep_time_interval=None.'
      )
    if self.read_only and self.keep_period is not None:
      self.keep_period = None
      self.should_keep_fn = None
      logging.warning(
          'CheckpointManagerOptions.read_only=True, setting keep_period=None'
          ' and should_keep_fn=None.'
      )
    if self.read_only and self.create:
      self.create = False
      logging.warning(
          'CheckpointManagerOptions.read_only=True, setting create=False.'
      )
    if self.read_only and self.cleanup_tmp_directories:
      self.cleanup_tmp_directories = False
      logging.warning(
          'CheckpointManagerOptions.read_only=True, setting'
          ' cleanup_tmp_directories=False.'
      )
    if self.read_only and self.save_on_steps:
      self.save_on_steps = None
      logging.warning(
          'CheckpointManagerOptions.read_only=True, setting save_on_steps=None.'
      )
    if self.read_only and self.todelete_subdir is not None:
      self.todelete_subdir = None
      logging.warning(
          'CheckpointManagerOptions.read_only=True, setting'
          ' todelete_subdir=None.'
      )
    if self.read_only and self.should_save_fn is not None:
      self.should_save_fn = None
      logging.warning(
          'CheckpointManagerOptions.read_only=True, setting'
          ' should_save_fn=None.'
      )
    if self.preservation_policy is None and self.should_keep_fn is not None:
      logging.warning(
          'CheckpointManagerOptions.should_keep_fn is set, setting'
          ' keep_period=None (was %s).',
          self.keep_period,
      )
      self.keep_period = None
    self.save_on_steps = frozenset(self.save_on_steps or ())


def _get_args_for_key(
    handler: CheckpointHandler, item_name: str
) -> Tuple[Type[CheckpointArgs], Type[CheckpointArgs]]:
  """Returns the (save, restore) args for the given item name."""

  if not isinstance(handler, CompositeCheckpointHandler):
    raise ValueError(
        'Expected handler to be a `CompositeCheckpointHandler`, but got'
        f' {type(handler)}.'
    )

  registry = handler._handler_registry  # pylint: disable=protected-access
  for (registered_item, _), handler in registry.get_all_entries().items():
    if registered_item == item_name:
      return checkpoint_args.get_registered_args_cls(handler)
  raise ValueError(f'Unknown key "{item_name}" in CompositeCheckpointHandler.')


def _create_root_directory(
    directory: epath.PathLike,
    multiprocessing_options: MultiprocessingOptions,
    file_options: Optional[FileOptions] = None,
) -> None:
  """Creates the top-level directory if it does not already exist."""
  if multiprocessing_options.active_processes is not None:
    raise NotImplementedError(
        'Option `create=True` with `active_processes` set is not'
        ' supported. Please create the root directory yourself.'
    )
  directory = epath.Path(directory)
  if not directory.exists() and utils.is_primary_host(
      multiprocessing_options.primary_host
  ):
    # exists_ok=True is required, see b/362903314.
    directory.mkdir(parents=True, exist_ok=True)
    logging.info('Created directory=%s', directory)
  multihost.sync_global_processes(
      multihost.unique_barrier_key(
          'CheckpointManager:create_directory',
          prefix=multiprocessing_options.barrier_sync_key_prefix,
          # suffix=None,
      ),
      timeout=multihost.DIRECTORY_CREATION_TIMEOUT,
      processes=multiprocessing_options.active_processes,
  )


def determine_default_item_mode_from_args(
    args: args_lib.CheckpointArgs,
) -> bool:
  if isinstance(args, args_lib.Composite):
    return False
  else:
    return True


def _determine_default_item_mode_from_directory(step_path: epath.Path) -> bool:
  return (step_path / DEFAULT_ITEM_NAME).exists()


class CheckpointManager(AbstractCheckpointManager, epy.ContextManager):
  """A generic, synchronous AbstractCheckpointManager implementation."""

  def __init__(
      self,
      directory: epath.PathLike,
      checkpointers: Optional[
          Union[AbstractCheckpointer, CheckpointersDict]
      ] = None,
      options: Optional[CheckpointManagerOptions] = None,
      metadata: Optional[Mapping[str, Any]] = None,
      item_names: Optional[Sequence[str]] = None,
      item_handlers: Optional[
          Union[CheckpointHandler, CheckpointHandlersDict]
      ] = None,
      logger: Optional[abstract_logger.AbstractLogger] = None,
      handler_registry: Optional[CheckpointHandlerRegistry] = None,
  ):
    """CheckpointManager constructor.

    IMPORTANT: `CheckpointManager` has been refactored to provide a new API.
    Please ensure you have migrated all existing use cases to the newer style by
    August 1st, 2024. Please see
    https://orbax.readthedocs.io/en/latest/guides/checkpoint/api_refactor.html
    for technical details.

    The `CheckpointManager` is ultimately backed by a single
    :py:class:`.Checkpointer`, to which saving and restoring is delegated.
    Behind
    step management options, metrics-related logic, and other frills, saving and
    restoring with `CheckpointManager` is quite similar to using
    `Checkpointer(CompositeCheckpointHandler)`.

    Example::

      # Multiple items.
      with CheckpointManager(
        'path/to/dir/',
        metadata={'version': 1.1, 'lang': 'en'},
      ) as mngr:
        mngr.save(0, args=args.Composite(
            train_state=args.StandardSave(train_state),
            custom_metadata=args.JsonSave(custom_metadata),
          )
        )
        restored = mngr.restore(0)
        print(restored.train_state)
        print(restored.custom_metadata)
        restored = mngr.restore(0, args=args.Composite(
            train_state=args.StandardRestore(abstract_train_state),
          )
        )
        print(restored.train_state)
        print(restored.custom_metadata)  # Error, not restored

      # Single, unnamed (default) item.
      with CheckpointManager(
          'path/to/dir/',
          options = CheckpointManagerOptions(max_to_keep=5, ...),
        ) as mngr:
        mngr.save(0, args=StandardSave(train_state))
        train_state = mngr.restore(0)
        train_state = mngr.restore(0,
        args=StandardRestore(abstract_train_state))

    IMPORTANT: Don't forget to use the keyword `args=...` for save and restore!
    Otherwise you will get the legacy API. This will not be necessary forever,
    but only until the legacy API is removed.

    IMPORTANT: The CheckpointManager is designed to be used as a context
    manager. Use `with CheckpointManager` schematic for automatic cleanup. If
    you can't use a context manager, always call `close()` to release resources
    properly.  Otherwise, background operations such as deleting old checkpoints
    might not finish before your program exits.

    CheckpointManager:
      - is NOT thread-safe.
      - IS multi-process-safe.
      - is NOT multi-job-safe.
    This means that CheckpointManager is intended to be created and called
    across all processes within a single job, where each process is
    single-threaded, but is not safe to use when multiple jobs each have
    CheckpointManager instances pointing to the same root directory. Concretely,
    this means that if you have a trainer job and one or more evaluator jobs,
    the CheckpointManager should be created and called across all processes
    in the trainer, but a CheckpointManager cannot be used in the evaluators.
    Instead, utilities used during evaluation can be found in
    `checkpoint_utils` (see
    https://orbax.readthedocs.io/en/latest/api_reference/checkpoint.checkpoint_utils.html).

    Args:
      directory: the top level directory in which to save all files.
      checkpointers: deprecated, do not use. use `handler_registry` instead.
      options: CheckpointManagerOptions. May be provided to specify additional
        arguments. If None, uses default values of CheckpointManagerOptions.
      metadata: High-level metadata that does not depend on step number. If
        `directory` is write enabled then given metadata is saved only once. A
        new CheckpointManager instance with that `directory` does not overwrite
        the existing metadata and ignores the current given metadata. If
        `directory` is read-only then the current given metadata is not saved as
        expected. A CheckpointManager instance with a read-only `directory` uses
        the metadata if already present, otherwise always uses the current given
        metadata.
      item_names: deprecated, do not use. use `handler_registry` instead.
      item_handlers: deprecated, do not use. use `handler_registry` instead.
      logger: A logger to log checkpointing events.
      handler_registry: A registry of handlers to use for checkpointing. This
        option is mutually exclusive with `checkpointers`,`item_handlers`, and
        'item_names'. See :py:class:`CheckpointHandlerRegistry` for more
        details.
    """
    jax.monitoring.record_event('/jax/orbax/checkpoint_manager/init')
    logging.info(
        '[process=%s][thread=%s] CheckpointManager init: checkpointers=%s,'
        ' item_names=%s, item_handlers=%s, handler_registry=%s',
        multihost.process_index(),
        threading.current_thread().name,
        checkpointers,
        item_names,
        item_handlers,
        handler_registry,
    )

    self._options = options or CheckpointManagerOptions()
    self._multiprocessing_options = self._options.multiprocessing_options
    self._save_decision_policy = (
        self._options.save_decision_policy
        or _get_default_save_decision_policy(self._options)
    )

    if self._options.preservation_policy and (
        self._options.keep_time_interval
        or self._options.max_to_keep
        or self._options.keep_period
        or self._options.should_keep_fn
    ):
      raise ValueError(
          '`preservation_policy` and `delete options` are mutually exclusive'
          ' - do not use together.'
      )

    self._preservation_policy = (
        self._options.preservation_policy
        or _get_default_preservation_policy(self._options)
    )

    if self._options.best_mode not in ['min', 'max']:
      raise ValueError('`best_mode` must be one of: "min", "max"')

    self._logger = logger or standard_logger.StandardLogger()

    if checkpointers and item_names:
      raise ValueError(
          '`item_names` and `checkpointers` are mutually exclusive - do not use'
          ' together.'
      )
    if checkpointers and item_handlers:
      raise ValueError(
          '`item_handlers` and `checkpointers` are mutually exclusive - do not'
          ' use together.'
      )
    if item_names and isinstance(item_handlers, CheckpointHandler):
      raise ValueError(
          '`item_handlers` in default item mode and `item_names` should not be'
          ' provided together.'
      )
    if checkpointers is not None and handler_registry is not None:
      raise ValueError(
          'Deprecated `checkpointers` can not be used with `handler_registry`.'
          ' Please follow the instructions at'
          ' https://orbax.readthedocs.io/en/latest/api_refactor.html to'
          ' migrate.'
      )

    if item_handlers is not None and handler_registry is not None:
      raise ValueError(
          '`item_handlers` and `handler_registry` are mutually exclusive -'
          ' prefer configuring the handler registry.'
      )

    if item_names is not None and handler_registry is not None:
      raise ValueError(
          '`item_names` and `handler_registry` are mutually exclusive - prefer'
          ' configuring the handler registry.'
      )

    # For async_checkpointer.
    self._non_blocking_metadata_store = checkpoint.metadata_store(
        enable_write=True
    )
    # For metadata checkpointer and regular checkpointer.
    self._blocking_metadata_store = checkpoint.metadata_store(
        enable_write=True, blocking_write=True
    )

    self._default_item = threading_lib.OptionalRef[bool]()
    if checkpointers is not None:
      jax.monitoring.record_event(
          '/jax/orbax/deprecation/checkpoint_manager_legacy_init'
      )
      logging.warning(
          'Configured `CheckpointManager` using deprecated legacy API. Please'
          ' follow the instructions at'
          ' https://orbax.readthedocs.io/en/latest/api_refactor.html to'
          ' migrate.'
      )
      self._default_item.set(isinstance(checkpointers, AbstractCheckpointer))
      self._checkpointer = self._configure_checkpointer_legacy_init(
          checkpointers, self._options
      )
    elif handler_registry is not None:
      # There is no way to know if this is a single, unnamed (default) item or
      # not, detemine this lazily instead on the first call to `save`, `restore`
      # or `item_metadata`. Once locked-in, the value of `_default_item` will
      # not change.
      self._default_item.set(None)
      self._checkpointer = self._configure_checkpointer_from_handler_registry(
          handler_registry,
          self._options,
      )
    elif item_names is None and item_handlers is None:
      # In this case, we can just default construct the
      # CheckpointHandlerRegistry and allow the user to lazily specify default
      # vs. multi-item mode.
      self._default_item.set(None)
      handler_registry = handler_registration.DefaultCheckpointHandlerRegistry()
      self._checkpointer = self._configure_checkpointer_from_handler_registry(
          handler_registry,
          self._options,
      )
    else:
      default_item = isinstance(item_handlers, CheckpointHandler)
      self._default_item.set(default_item)
      self._checkpointer = (
          self._configure_checkpointer_from_item_names_and_handlers(
              item_names,
              item_handlers,
              self._options,
              default_item,
          )
      )

    if (
        self._options.async_options is not None
        and self._options.async_options.post_finalization_callback is not None
        and not isinstance(self._checkpointer, AsyncCheckpointer)
    ):
      raise ValueError(
          'AsyncOptions.post_finalization_callback is only supported with'
          ' AsyncCheckpointer. But final resolved checkpointer is: '
          f' {self._checkpointer}'
      )

    self._directory = epath.Path(directory)
    self._save_tracker = synchronization.OpTrackerFactory.create_tracker(
        'checkpoint_manager_save'
    )
    if self._options.read_only:
      logging.warning('Given directory is read only=%s', self._directory)
    if self._options.create:
      _create_root_directory(
          self._directory,
          self._multiprocessing_options,
          self._options.file_options,
      )


    # Cleanup directories from previous runs that may not have been finalized.
    if self._options.cleanup_tmp_directories:
      self._cleanup_tmp_directories()

    self._step_name_format = (
        self._options.step_name_format
        or step_lib.standard_name_format(
            step_prefix=self._options.step_prefix,
            step_format_fixed_length=self._options.step_format_fixed_length,
            single_host_load_and_broadcast=(
                self._options.single_host_load_and_broadcast
            ),
        )
    )

    self._checkpoints = checkpoint_info.CheckpointInfos(
        self._load_checkpoint_infos()
    )

    self._metadata_dir = self.directory / METADATA_ITEM_NAME
    if self._options.read_only and not self._metadata_dir.exists():
      custom_metadata = {} if metadata is None else dict(metadata)
    else:
      custom_metadata = None
    self._root_metadata = threading_lib.Ref(
        RootMetadata(
            custom_metadata=custom_metadata,
        )
    )

    self._maybe_save_root_metadata(metadata)

    # TODO: b/359854428 - Move Finalize biz logic to a separate class/module.
    self._finalize_thread = threading_lib.OptionalRef[_FinalizeThread]()

    self._checkpoint_deleter: deleter.CheckpointDeleter = (
        deleter.create_checkpoint_deleter(
            self._multiprocessing_options.primary_host,
            self._directory,
            self._options.todelete_subdir,
            self._step_name_format,
            self._options.enable_hns_rmtree,
            self._options.enable_background_delete,
        )
    )

    logging.info(
        '[process=%s][thread=%s] CheckpointManager created,  primary_host=%s,'
        ' CheckpointManagerOptions=%s, root_directory=%s: %s',
        multihost.process_index(),
        threading.current_thread().name,
        self._multiprocessing_options.primary_host,
        self._options,
        self.directory,
        self,
    )

  def _create_thread_safe_barrier_sync_fn(self) -> Callable[[str], None]:
    """Returns a barrier sync function to be called from threads.

    The function accepts a key, but the timeout is already set up using
    `AsyncOptions.timeout_secs` attribute.

    The Jax based barrier sync util, `sync_global_devices`, should not be called
    concurrently. Otherwise, it may cause a deadlock.

    In general, any Jax function with `collectives` should not be called
    concurrently to avoid deadlocks.
    """
    async_options = self._options.async_options or AsyncOptions()
    timeout_secs = async_options.timeout_secs
    barrier_sync_fn = (
        async_options.barrier_sync_fn
        or multihost.get_barrier_sync_fn(
            processes=self._multiprocessing_options.active_processes
        )
    )
    return lambda key: barrier_sync_fn(key=key, timeout_ms=timeout_secs * 1000)

  def _configure_checkpointer_common(
      self,
      handler: CompositeCheckpointHandler,
      options: CheckpointManagerOptions,
      use_async: bool,
  ) -> Checkpointer:
    if use_async:
      return async_checkpointer.AsyncCheckpointer(
          handler,
          multiprocessing_options=options.multiprocessing_options,
          async_options=options.async_options or AsyncOptions(),
          file_options=options.file_options,
          checkpoint_metadata_store=self._non_blocking_metadata_store,
          temporary_path_class=options.temporary_path_class,
      )
    else:
      return Checkpointer(
          handler,
          multiprocessing_options=options.multiprocessing_options,
          file_options=options.file_options,
          checkpoint_metadata_store=self._blocking_metadata_store,
          temporary_path_class=options.temporary_path_class,
      )

  def _configure_checkpointer_legacy_init(
      self,
      checkpointers: Union[AbstractCheckpointer, CheckpointersDict],
      options: CheckpointManagerOptions,
  ) -> Checkpointer:
    """Initializes _CompositeCheckpointer with legacy style checkpointers."""
    if self._multiprocessing_options.primary_host != 0:
      raise ValueError(
          f'`primary_host`={self._multiprocessing_options.primary_host} is not'
          ' supported in legacy API.'
      )

    item_handlers = {}
    if isinstance(checkpointers, Checkpointer):
      use_async = is_async_checkpointer(checkpointers)
      if isinstance(checkpointers, async_checkpointer.AsyncCheckpointer):
        async_timeout = checkpointers._async_manager._timeout_secs  # pylint: disable=protected-access
      else:
        async_timeout = None
      item_handlers[DEFAULT_ITEM_NAME] = checkpointers.handler
    elif isinstance(checkpointers, dict):
      individual_use_async = []
      async_timeout = 0
      for item_name, checkpointer in checkpointers.items():
        if not isinstance(checkpointer, Checkpointer):
          raise ValueError(
              f'Value corresponding to {item_name} in `checkpointers` is not a'
              f' Checkpointer. Found {type(checkpointer)}.'
          )
        individual_use_async.append(is_async_checkpointer(checkpointer))
        if isinstance(checkpointer, async_checkpointer.AsyncCheckpointer):
          async_timeout = max(
              async_timeout, checkpointer._async_manager._timeout_secs  # pylint: disable=protected-access
          )
        if item_name in RESERVED_ITEM_NAMES:
          raise ValueError(
              f'Found {item_name} in `checkpointers`; this is a reserved key.'
          )
        item_handlers[item_name] = checkpointer.handler
      if any(individual_use_async) and not all(individual_use_async):
        logging.error(
            'Orbax `CheckpointManager` is transitioning toward using'
            ' asynchronous saving logic under the hood in all cases. Users that'
            ' configure `CheckpointManager` with some `Checkpointer`s and some'
            ' `AsyncCheckpointer`s will now see asynchronous logic used to save'
            ' all items. This may result in breakages if the code is assuming'
            ' that certain objects will be available immediately after saving.'
            ' Ensure that if you depend on the result of `save` being fully'
            ' written at a particular moment, use `wait_until_finished()`.'
        )
      use_async = any(individual_use_async)
      async_timeout = async_timeout or None
    else:
      raise ValueError(
          f'Invalid type for `checkpointers`. Found {checkpointers}.'
      )

    item_handlers[METRIC_ITEM_NAME] = self._metrics_handler
    if options.async_options is None:
      options.async_options = (
          AsyncOptions(timeout_secs=async_timeout)
          if async_timeout is not None
          else AsyncOptions()
      )
    return self._configure_checkpointer_common(
        CompositeCheckpointHandler(
            composite_options=composite_checkpoint_handler.CompositeOptions(
                multiprocessing_options=options.multiprocessing_options,
                file_options=options.file_options,
                async_options=options.async_options,
            ),
            **item_handlers,
        ),
        options,
        use_async,
    )


  def _configure_checkpointer_from_item_names_and_handlers(
      self,
      item_names: Optional[Sequence[str]],
      item_handlers: Optional[Union[CheckpointHandler, CheckpointHandlersDict]],
      options: CheckpointManagerOptions,
      default_item: bool,
  ) -> Checkpointer:
    """Initializes _CompositeCheckpointer given `item_names`."""
    if (
        self._multiprocessing_options.primary_host is None
        and item_handlers is None
    ):
      raise ValueError(
          'When primary_host is set to None, item_handlers must be provided to'
          ' match with the primary_host setting.'
      )
    if default_item:
      item_handler = (
          item_handlers
          if isinstance(item_handlers, CheckpointHandler)
          else None
      )
      all_item_handlers = {DEFAULT_ITEM_NAME: item_handler}
    else:
      # Initialize all_item_handlers with None or empty.
      if item_names:
        all_item_handlers = {item_name: None for item_name in item_names}
      else:
        all_item_handlers = {}
      # Update all_item_handlers with provided CheckpointHandlers.
      if item_handlers and isinstance(item_handlers, Mapping):
        for item_name, handler in item_handlers.items():
          all_item_handlers[item_name] = handler

    for item_name in all_item_handlers:
      if item_name in RESERVED_ITEM_NAMES:
        raise ValueError(
            f'Found {item_name} in `checkpointers`; this is a reserved key.'
        )
    all_item_handlers[METRIC_ITEM_NAME] = self._metrics_handler
    # CompositeCheckpointHandler defers per-item handler creation until
    # save/restore time.
    async_options = options.async_options or AsyncOptions()
    async_options.create_directories_asynchronously &= (
        options.enable_async_checkpointing
    )
    return self._configure_checkpointer_common(
        CompositeCheckpointHandler(
            composite_options=composite_checkpoint_handler.CompositeOptions(
                multiprocessing_options=options.multiprocessing_options,
                file_options=options.file_options,
                async_options=async_options,
            ),
            **all_item_handlers,
        ),
        options,
        options.enable_async_checkpointing,
    )

  def _configure_checkpointer_from_handler_registry(
      self,
      handler_registry: CheckpointHandlerRegistry,
      options: CheckpointManagerOptions,
  ) -> Checkpointer:
    """Initializes _CompositeCheckpointer given a `handler_registry`."""
    metrics_handler = self._metrics_handler
    if not handler_registry.has(METRIC_ITEM_NAME, args_lib.JsonSave):
      handler_registry.add(METRIC_ITEM_NAME, args_lib.JsonSave, metrics_handler)
    if not handler_registry.has(METRIC_ITEM_NAME, args_lib.JsonRestore):
      handler_registry.add(
          METRIC_ITEM_NAME, args_lib.JsonRestore, metrics_handler
      )

    # CompositeCheckpointHandler defers per-item handler creation until
    # save/restore time.
    async_options = options.async_options or AsyncOptions()
    async_options.create_directories_asynchronously &= (
        options.enable_async_checkpointing
    )
    return self._configure_checkpointer_common(
        CompositeCheckpointHandler(
            composite_options=composite_checkpoint_handler.CompositeOptions(
                multiprocessing_options=options.multiprocessing_options,
                async_options=async_options,
            ),
            handler_registry=handler_registry,
        ),
        options,
        options.enable_async_checkpointing,
    )

  @property
  def directory(self) -> epath.Path:
    """Returns the top-level directory containing checkpoints for all items."""
    return self._directory

  def all_steps(self, read: bool = False) -> Sequence[int]:
    """Returns all steps tracked by the manager.

    Args:
      read: If True, forces a read directly from the storage location.
        Otherwise, a cached result can be returned.

    Returns:
      A sequence of steps (integers)
    """
    if read:
      logging.warning(
          '`read` option is deprecated. Use `reload` to read from disk.'
      )
      self._checkpoints.set(self._load_checkpoint_infos())
    return [ckpt.step for ckpt in self._checkpoints]

  def latest_step(self) -> Optional[int]:
    """Returns the latest step saved.

    Returns None if no steps have been saved.

    Returns:
      A step (int) or None if no steps are present.
    """
    latest = self._checkpoints.latest()
    return latest.step if latest else None

  def best_step(self) -> Optional[int]:
    """Returns the best step saved, as defined by `options.best_fn`.

    Returns None if no steps have been saved.

    Returns:
      A step (int) or None if no steps are present.
    """
    if not self._track_best:
      return self.latest_step()
    if self._checkpoints.empty():
      return None
    _, sorted_checkpoints = self._sort_checkpoints_by_metrics(self._checkpoints)
    if not sorted_checkpoints:
      return None
    return sorted_checkpoints[-1].step

  def reload(self):
    """Reloads internal properties.

    Resets internal cache of checkpoint steps, in case the directory managed
    by this object has been updated externally.
    """
    self._checkpoints.set(self._load_checkpoint_infos())

  def reached_preemption(self, step: int) -> bool:
    """Returns True if a preemption sync point has been reached."""
    return utils.reached_preemption(step)

  def should_save(self, step: int) -> bool:
    """Returns True if a checkpoint should be saved for the current step.

    This depends the previous step and SaveDecisionPolicy.

    Args:
      step: int

    Returns:
      True if the checkpoint should be saved.
    """
    if self._options.read_only:
      logging.warning('%s is read only, save will be skipped', self.directory)
      return False
    last_checkpoint_step = self.latest_step()
    # Ensure current step is between the last step and next step (accounting for
    # save interval).
    if last_checkpoint_step is not None and last_checkpoint_step >= step:
      return False

    if self._options.enable_should_save_is_saving_in_progress_check:
      is_saving_in_progress = self.is_saving_in_progress()
    else:
      is_saving_in_progress = False
    reached_preemption = self.reached_preemption(step)
    current_step_info = checkpoint_info.CheckpointInfo(
        step=step,
        time=datetime.datetime.now(tz=datetime.timezone.utc),
        metrics=None,
    )
    context = save_decision_policy_lib.DecisionContext(
        is_saving_in_progress=is_saving_in_progress,
        reached_preemption=reached_preemption,
        multiprocessing_options=self._multiprocessing_options,
    )
    return self._save_decision_policy.should_save(
        current_step_info, previous_steps=self._checkpoints, context=context
    )

  def _get_save_directory(
      self,
      step: int,
      directory: epath.Path,
  ) -> epath.Path:
    """Returns the standardized path to a save directory for a single item."""

    return step_lib.build_step_path(directory, self._step_name_format, step)

  def _get_write_step_directory(
      self, step: int, root_dir: epath.Path
  ) -> epath.Path:
    return self._get_save_directory(step, root_dir)

  def _get_read_step_directory(
      self, step: int, root_dir: epath.Path
  ) -> epath.Path:
    if self._options.step_name_format is not None:
      return self._options.step_name_format.find_step(root_dir, step).path
    else:
      return self._get_save_directory(step, root_dir)

  def delete(self, step: int):
    """See superclass documentation.

    Delete can be run asynchronously if
    CheckpointManagerOptions.enable_background_delete is set to True.

    Args:
      step: The step to delete.

    Raises:
      FileNotFoundError: If the step does not exist.
    """
    if self._options.read_only:
      logging.warning('%s is read only, delete will be skipped', self.directory)
      return
    if step not in self.all_steps():
      raise FileNotFoundError(
          f'Requested deleting a non-existent step: {step}.'
      )
    self._checkpoint_deleter.delete(step)
    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'CheckpointManager:deleted_step',
            prefix=self._multiprocessing_options.barrier_sync_key_prefix,
            # suffix=str(step),
        ),
        timeout=multihost.DIRECTORY_DELETION_TIMEOUT,
        processes=self._multiprocessing_options.active_processes,
    )
    self._checkpoints.delete_if(lambda info: info.step == step)

  def _validate_args(
      self,
      items: Optional[Union[Any, Mapping[str, Any]]],
      args: Optional[args_lib.CheckpointArgs],
  ):
    if isinstance(items, args_lib.CheckpointArgs):
      raise ValueError(
          'Found an instance of `CheckpointArgs` provided for `items`. This may'
          ' be due to misuse of the newer API - make sure to specify the'
          ' argument keyword (e.g. `args=args`).'
      )
    if args is not None:
      if not isinstance(args, args_lib.CheckpointArgs):
        raise ValueError(
            f'Expected args of type `CheckpointArgs`; found {type(args)}.'
        )
      if self._default_item.get():
        if isinstance(args, args_lib.Composite):
          raise ValueError(
              'Cannot provide `args` of type `Composite` when dealing with a'
              ' single, unnamed (default) checkpointable object.'
          )
      else:
        if not isinstance(args, args_lib.Composite):
          raise ValueError(
              'Must provide `args` of type `Composite` when dealing with'
              ' multiple checkpointable objects.'
          )

  def save(
      self,
      step: int,
      items: Optional[Union[Any, Mapping[str, Any]]] = None,
      save_kwargs: Optional[Union[SaveParams, Mapping[str, SaveParams]]] = None,
      metrics: Optional[PyTree] = None,
      force: Optional[bool] = False,
      args: Optional[args_lib.CheckpointArgs] = None,
      custom_metadata: dict[str, Any] | None = None,
  ) -> bool:
    """Saves the provided items.

    This method should be called by all hosts - process synchronization and
    actions that need to be performed on only one host are managed internally.

    NOTE: The `items` and `save_kwargs` arguments are deprecated, use `args`
    instead. Make sure to configure `CheckpointManager` with `item_names`.

    `args` should be a subclass of
    `orbax.checkpoint.args.CheckpointArgs`, the specific type of which is used
    to indicate what logic is used to save the object. For a typical, PyTree of
    arrays, use `StandardSave`/`StandardRestore`.

    When constructing the `CheckpointManager`, if no `item_names` were provided,
    it is assumed that we are managing a single object. If `item_names` were
    provided, it is assumed that we are managing multiple objects, and `args`
    must be `orbax.checkpoint.args.CompositeArgs`. See below for details.

    Example::

      # Single item
      mngr = ocp.CheckpointManager(directory)
      mngr.save(step, args=ocp.args.StandardSave(my_train_state))

      # Multiple items
      mngr = ocp.CheckpointManager(directory, item_names=('state', 'meta'))
      mngr.save(step, args=ocp.args.Composite(
          state=ocp.args.StandardSave(my_train_state),
          meta=ocp.args.JsonSave(my_metadata)
      ))

    Args:
      step: current step, int
      items: a savable object, or a dictionary of object name to savable object.
      save_kwargs: save kwargs for a single Checkpointer, or a dictionary of
        object name to kwargs needed by the Checkpointer implementation to save
        the object.
      metrics: a dictionary of metric name (string) to numeric value to be
        tracked along with this checkpoint. Required if `options.best_fn` is
        set. Allows users to specify a metric value to determine which
        checkpoints are best and should be kept (in conjunction with
        `options.max_to_keep`).
      force: if `True`, this method will attempt to save a checkpoint regardless
        of the result of `AbstractCheckpointManager.should_save(step)`. By
        default, `save` will only write a checkpoint to disk when the options
        permit, e.g. when `step` is in `options.save_interval_steps` or
        `options.save_on_steps`. Setting `force=True` will not overwrite
        existing checkpoints.
      args: `CheckpointArgs` which is used to save checkpointable objects with
        the appropriate logic.
      custom_metadata: a dictionary of custom metadata to be written to the
        checkpoint directory via StepMetadata.

    Returns:
      bool indicating whether a save operation was performed.
    Raises:
      ValueError: if `track_best` was indicated but `metrics` is not provided.
      ValueError: directory creation failed.
      ValueError: if an item is provided for which no `Checkpointer` is
      found.
      ValueError: if the checkpoint already exists.
    """
    process_index = multihost.process_index()
    step_stats = step_statistics.SaveStepStatistics()
    step_stats.step = step
    step_stats.checkpoint_manager_blocking_start_time = time.time()
    step_stats.directory = str(self.directory)

    if items is None and args is None:
      raise ValueError('Must provide `args` for `save`.')
    self._default_item.set_if_none(determine_default_item_mode_from_args(args))
    self._validate_args(items, args)
    if not force and not self.should_save(step):
      return False

    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'CheckpointManager:save_start',
            prefix=self._multiprocessing_options.barrier_sync_key_prefix,
        ),
        processes=self._multiprocessing_options.active_processes,
        record_event_name=(
            '/jax/orbax/write/checkpoint_start_sync_duration_secs'
        ),
    )
    if self.reached_preemption(step):
      logging.info(
          '[process=%s] Saving checkpoint at step %d due to preemption.',
          process_index,
          step,
      )
      step_stats.reached_preemption = True
      step_stats.preemption_received_at = time.time()

    # Wait for ongoing saves to complete. Only applicable if some of the
    # checkpointers are AsyncCheckpointers.
    # Must happen after `should_save` to avoid blocking callers.
    step_stats.wait_for_prev_start_time = time.time()
    self.wait_until_finished()
    step_stats.wait_for_prev_duration_secs = (
        time.time() - step_stats.wait_for_prev_start_time
    )

    jax.monitoring.record_event_duration_secs(
        '/jax/checkpoint/write/wait_for_prev_duration_secs',
        step_stats.wait_for_prev_duration_secs,
    )
    self._save_tracker = synchronization.OpTrackerFactory.create_tracker(
        'checkpoint_manager_save'
    )
    self._save_tracker.start()
    if step in self.all_steps():
      raise StepAlreadyExistsError(
          f'Checkpoint for step {step} already exists.'
      )

    if items is None:
      items = {}
    if save_kwargs is None:
      save_kwargs = {}
    if self._default_item.get():
      items = {DEFAULT_ITEM_NAME: items}
      save_kwargs = {DEFAULT_ITEM_NAME: save_kwargs}

    if self._track_best and metrics is None:
      logging.warning('Requested `tracked_metric`; did not provide metrics.')

    if args is None:
      args_dict = {}
      for key, item in items.items():
        save_ckpt_arg_cls, _ = _get_args_for_key(
            self._checkpointer.handler,
            key,
        )
        extra_args = save_kwargs[key] if key in save_kwargs else {}
        extra_args = extra_args or {}
        args_dict[key] = save_ckpt_arg_cls(item, **extra_args)  # pylint: disable=too-many-function-args  # pytype: disable=wrong-arg-count
      args = args_lib.Composite(**args_dict)
    else:
      if self._default_item.get():
        args = args_lib.Composite(**{DEFAULT_ITEM_NAME: args})
      else:
        if not isinstance(args, args_lib.Composite):
          raise ValueError(
              f'Expected args of type `Composite`; found {type(args)}.'
          )
        args = typing.cast(args_lib.Composite, args)

    args_dict = dict(args.items())
    if any(
        [item_name in RESERVED_ITEM_NAMES for item_name in args_dict.keys()]
    ):
      raise ValueError(
          'Some provided items have prohibited reserved names:'
          f' {args_dict.keys()}. Reserved names: {RESERVED_ITEM_NAMES}.'
      )
    if (
        metrics is not None and self._track_best
    ) and not self._options.prevent_write_metrics:
      args_dict[METRIC_ITEM_NAME] = args_lib.JsonSave(metrics)
    args = args_lib.Composite(**args_dict)

    save_directory = self._get_write_step_directory(step, self.directory)
    logging.info(
        '[process=%s] Saving checkpoint at step %d', process_index, step
    )
    step_stats.checkpointer_blocking_start_time = time.time()
    self._checkpointer.save(
        save_directory, args=args, custom_metadata=custom_metadata, force=True
    )
    step_stats.checkpointer_blocking_duration_secs = (
        time.time() - step_stats.checkpointer_blocking_start_time
    )

    self._add_checkpoint_info(step, metrics)
    step_stats.get_old_steps_start_time = time.time()
    steps_to_remove = self._get_old_steps_to_remove()
    step_stats.get_old_steps_duration_secs = (
        time.time() - step_stats.get_old_steps_start_time
    )

    jax.monitoring.record_event_duration_secs(
        '/jax/checkpoint/write/get_old_steps_duration_secs',
        step_stats.get_old_steps_duration_secs,
    )

    self._checkpoints.delete_if(lambda info: info.step in steps_to_remove)
    # Sync needed to ensure that old steps to remove are retrieved before
    # actually deleting them during finalize, since retrieval can involve
    # looking at the directory.
    multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'CheckpointManager:old_steps_to_remove',
            prefix=self._multiprocessing_options.barrier_sync_key_prefix,
            # suffix=str(step),
        ),
        processes=self._multiprocessing_options.active_processes,
    )

    current_thread = threading.current_thread()
    if is_async_checkpointer(self._checkpointer):

      def launch_finalize_thread() -> _FinalizeThread:
        is_finalize_in_progress = self._finalize_thread.map(
            lambda t: t is not None and t.is_alive()
        )
        assert not is_finalize_in_progress, (
            'Save finalization already in progress for'
            f' step={self._finalize_thread.get_not_none().step()}'
        )
        finalize_thread_name = 'save_finalize'
        logging.info(
            '[process=%s][thread=%s][step=%s] Starting CheckpointManager Save'
            ' Finalize thread=%s',
            process_index,
            current_thread.name,
            step,
            finalize_thread_name,
        )
        finalize_thread = _FinalizeThread(
            step=step,
            name=finalize_thread_name,
            target=self._finalize,
            args=(step, steps_to_remove),
        )
        finalize_thread.start()
        return finalize_thread

      self._finalize_thread.set_from(launch_finalize_thread)

    else:
      self._finalize(step, steps_to_remove)
      logging.info(
          '[process=%s][thread=%s][step=%s] Finished synchronous save.',
          process_index,
          current_thread.name,
          step,
      )

    step_stats.synchronous = not is_async_checkpointer(self._checkpointer)
    step_stats.checkpoint_manager_blocking_duration_secs = (
        time.time() - step_stats.checkpoint_manager_blocking_start_time
    )
    self._logger.log_entry(dataclasses.asdict(step_stats))
    return True

  def _maybe_get_default_item(
      self, composite_result: args_lib.Composite
  ) -> Union[Any, args_lib.Composite]:
    if self._default_item.get():
      if DEFAULT_ITEM_NAME not in composite_result:
        raise ValueError(
            'Unable to retrieve default item. Please ensure that a handler for'
            ' the default item is registered using `handler_registry` when'
            ' initializing the `CheckpointManager`.'
        )
      return composite_result[DEFAULT_ITEM_NAME]
    return composite_result

  def restore(
      self,
      step: Optional[int],
      items: Optional[Union[Any, Mapping[str, Any]]] = None,
      restore_kwargs: Optional[
          Union[RestoreParams, Mapping[str, RestoreParams]]
      ] = None,
      directory: Optional[epath.PathLike] = None,
      args: Optional[args_lib.CheckpointArgs] = None,
  ) -> Union[Any, Mapping[str, Any]]:
    """Restores from the given step and provided items.

    This method should be called by all hosts - process synchronization and
    actions that need to be performed on only one host are managed internally.

    NOTE: The `items` and `restore_kwargs` arguments are deprecated, use `args`
    instead. Make sure to configure `CheckpointManager` with `item_names`.
    See `save` docstring for additional details.

    Example::

      # Single item
      mngr = ocp.CheckpointManager(directory)
      mngr.restore(step, args=ocp.args.StandardRestore(abstract_train_state))

      # Multiple items
      mngr = ocp.CheckpointManager(directory, item_names=('state', 'meta'))
      mngr.restore(step, args=ocp.args.Composite(
          state=ocp.args.StandardRestore(abstract_train_state),
          meta=ocp.args.JsonRestore(),
      ))
      # If it is acceptable to restore without providing additional arguments,
      # and if a save has already been performed, it is ok to do the following:
      mngr.restore(step, args=ocp.args.Composite(state=None, meta=None))
      # If a save has not already been performed, there is no way for Orbax to
      # know how to restore the objects. If a save has already been performed,
      # it remembers the logic used to save the objects.

    Args:
      step: current step, int
      items: a restoreable object, or a dictionary of object name to restorable
        object.
      restore_kwargs: restore kwargs for a single Checkpointer, or a dictionary
        of object name to kwargs needed by the Checkpointer implementation to
        restore the object.
      directory: if provided, uses the given directory rather than the
        `directory` property of this class. Can be used to restore checkpoints
        from an independent location.
      args: `CheckpointArgs` which is used to restore checkpointable objects
        with the appropriate logic.

    Returns:
      If managing a single item, returns a single checkpointable object.
      If managing multiple items, returns ocp.args.Composite, where the keys
      are item names, and values are checkpointable objects.

    Raises:
      FileNotFoundError: If no steps are found in the directory.
    """
    if step is None:
      step = self.latest_step()
      if step is None:
        raise FileNotFoundError(f'No steps found in {self.directory}.')
    directory = directory or self.directory
    directory = epath.Path(directory)
    step_stats = step_statistics.RestoreStepStatistics()
    step_stats.step = step
    step_stats.checkpoint_manager_start_time = time.time()
    step_stats.directory = str(directory)

    self._default_item.set_if_none(
        _determine_default_item_mode_from_directory(
            self._get_read_step_directory(step, directory)
        )
    )
    self._validate_args(items, args)

    if items is None:
      items = {}
    elif self._default_item.get():
      items = {DEFAULT_ITEM_NAME: items}
    if restore_kwargs is None:
      restore_kwargs = {}
    elif self._default_item.get():
      restore_kwargs = {DEFAULT_ITEM_NAME: restore_kwargs}

    if args is None:
      args_dict = {}
      item_keys = set(items.keys()) | set(restore_kwargs.keys())
      for key in item_keys:
        _, restore_ckpt_arg_cls = _get_args_for_key(
            self._checkpointer.handler,
            key,
        )
        item = items[key] if key in items else None
        extra_args = restore_kwargs[key] if key in restore_kwargs else {}
        extra_args = extra_args or {}
        args_dict[key] = restore_ckpt_arg_cls(item, **extra_args)  # pylint: disable=too-many-function-args  # pytype: disable=wrong-arg-count
      args = args_lib.Composite(**args_dict)
    else:
      if self._default_item.get():
        args = args_lib.Composite(**{DEFAULT_ITEM_NAME: args})
      else:
        args = typing.cast(args_lib.Composite, args)

    restore_directory = self._get_read_step_directory(step, directory)
    step_stats.checkpointer_start_time = time.time()
    restored = self._checkpointer.restore(restore_directory, args=args)
    step_stats.checkpointer_duration_secs = (
        time.time() - step_stats.checkpointer_start_time
    )

    step_stats.checkpoint_manager_duration_secs = (
        time.time() - step_stats.checkpoint_manager_start_time
    )
    self._logger.log_entry(dataclasses.asdict(step_stats))

    return self._maybe_get_default_item(restored)

  def item_metadata(
      self, step: int
  ) -> Union[Any, args_lib.Composite, ItemMetadata]:
    """Retrieves metadata for all known items.

    Important note: This method will soon be deprecated in favor of
    `metadata().item_metadata`. Please use that method instead.

    Note that metadata will only be returned for items that can actually be
    interpreted. If an item is present in the checkpoint but not registered
    (using a prior save or restore, or with `handler_registry` at init), the
    item will not be returned.

    Args:
      step: The step to retrieve metadata for.

    Returns:
      Either metadata for the item itself, if in default-item mode, or a
      Composite of metadata for each item.
    """
    return self.metadata(step).item_metadata

  # TODO(b/370812224): Deprecate in favor of StepMetadata.metrics
  def metrics(self, step: int) -> Optional[PyTree]:
    return self._get_metrics(step)

  @property
  def _metrics_handler(self) -> CheckpointHandler:
    return JsonCheckpointHandler(
        filename=METRIC_ITEM_NAME,
        multiprocessing_options=self._multiprocessing_options,
    )

  @property
  def _track_best(self):
    """Returns true if we should track the best checkpoints by given metric."""
    return self._options.best_fn is not None

  def _load_checkpoint_infos(self) -> List[CheckpointInfo]:
    """Loads a list of CheckpointInfo for existing checkpoints.

    If none are present, returns empty list.

    Returns:
      a list of CheckpointInfo, sorted by increasing step.
    """
    if not self.directory.exists():
      return []
    start = time.time()
    step_metadatas = self._step_name_format.find_all(self.directory)

    def build_checkpoint_info(step_metadata):
      return CheckpointInfo(
          step=step_metadata.step,
          time=step_metadata.commit_timestamp,
          metrics=self.metrics(step_metadata.step),
      )

    with concurrent.futures.ThreadPoolExecutor() as executor:
      checkpoint_infos = list(
          executor.map(build_checkpoint_info, step_metadatas)
      )
      checkpoint_infos.sort(key=lambda x: x.step)  # Prefer in-place sort.
      jax.monitoring.record_event_duration_secs(
          '/jax/checkpoint/read/load_all_step_metadata_duration_secs',
          time.time() - start,
      )
      logging.info(
          'Found %d checkpoint steps in %s',
          len(checkpoint_infos),
          self.directory,
      )
      return checkpoint_infos

  def _add_checkpoint_info(self, step: int, metrics: Optional[PyTree]):
    self._checkpoints.append(
        CheckpointInfo(
            step, datetime.datetime.now(tz=datetime.timezone.utc), metrics
        )
    )

  def _root_metadata_file_path(self, legacy: bool = False) -> epath.Path:
    if not self._metadata_dir.exists():
      raise ValueError(
          f'Root Metadata directory is not initialized: {self._metadata_dir}'
      )
    return checkpoint.root_metadata_file_path(self._metadata_dir, legacy=legacy)

  def _maybe_save_root_metadata(
      self, custom_metadata: Mapping[str, Any] | None
  ):
    """Saves CheckpointManager level metadata, skips if already present."""
    if self._options.save_root_metadata:
      if (
          custom_metadata is not None
          and not self._options.read_only
          and utils.is_primary_host(self._multiprocessing_options.primary_host)
      ):
        self._metadata_dir.mkdir(parents=True, exist_ok=True)
        file_path = self._root_metadata_file_path()
        if not file_path.exists():  # May have been created by a previous run.
          metadata_to_save = self._root_metadata.get()
          if custom_metadata is not None:
            metadata_to_save.custom_metadata = dict(custom_metadata)
          self._blocking_metadata_store.write(
              file_path, serialize_root_metadata(metadata_to_save)
          )

      multihost.sync_global_processes(
          multihost.unique_barrier_key(
              'CheckpointManager:save_root_metadata',
              prefix=self._multiprocessing_options.barrier_sync_key_prefix,
          ),
          processes=self._multiprocessing_options.active_processes,
      )

  def _get_metrics(self, step: int) -> PyTree | None:
    try:
      # Use handler directly, since this happens in a background thread and
      # barriers cannot be used. This usage pattern is not
      # recommended in other contexts.
      metrics = self._metrics_handler.restore(
          self._get_read_step_directory(step, self.directory) / METRIC_ITEM_NAME
      )
      return metrics
    except FileNotFoundError:
      logging.log_first_n(
          logging.INFO,
          'Missing metrics for step %d',
          1,  # log only once
          step,
      )
      return None

  def _get_step_metadata(self, step: int) -> StepMetadata:
    step_metadata = self._checkpointer.metadata(
        self._get_read_step_directory(step, self.directory),
    )

    self._default_item.set_if_none(
        _determine_default_item_mode_from_directory(
            self._get_read_step_directory(step, self.directory)
        )
    )
    step_metadata.item_metadata = self._maybe_get_default_item(
        step_metadata.item_metadata
    )

    metrics = self._get_metrics(step)
    if metrics is not None:
      validated_metrics = step_metadata_serialization.deserialize(
          {}, metrics=dict(metrics)
      ).metrics
      step_metadata = dataclasses.replace(
          step_metadata,
          metrics=validated_metrics,
      )

    return step_metadata

  def _get_root_metadata(self) -> RootMetadata:
    if self._root_metadata.get().custom_metadata is None:
      if self._metadata_dir.exists():
        file_path = self._root_metadata_file_path()
        if not file_path.exists():
          logging.warning(
              'New root metadata file not found in directory: %s. '
              'Will try to read legacy file.',
              self._metadata_dir,
          )
          file_path = self._root_metadata_file_path(legacy=True)
        serialized_metadata = self._blocking_metadata_store.read(file_path)
        if serialized_metadata is None:
          raise IOError(f'Failed to read metadata from {file_path}')
        self._root_metadata.set(
            root_metadata_serialization.deserialize(serialized_metadata)
        )
      else:
        self._root_metadata.get().custom_metadata = {}
    return self._root_metadata.get()

  @overload
  def metadata(self, step: None = None) -> RootMetadata:
    ...

  @overload
  def metadata(self, step: int) -> StepMetadata:
    ...

  def metadata(self, step: int | None = None) -> RootMetadata | StepMetadata:
    """See superclass documentation."""
    if step is not None:
      return self._get_step_metadata(step)
    return self._get_root_metadata()

  def _sort_checkpoints_by_metrics(
      self, checkpoints: checkpoint_info.CheckpointInfos
  ) -> Tuple[List[CheckpointInfo], List[CheckpointInfo]]:
    """Sorts `checkpoints` in order of increasing metric quality.

    Checkpoints without corresponding metrics set will be at the beginning.

    Args:
      checkpoints: CheckpointInfos to sort.

    Returns:
      Tuple of CheckpointInfo lists:
      (checkpoints_without_metrics, checkpoints_sorted_by_metrics)
    """
    without_metrics = [info for info in checkpoints if info.metrics is None]
    with_metrics = [info for info in checkpoints if info.metrics is not None]

    return without_metrics, sorted(
        with_metrics,
        key=lambda info: self._options.best_fn(info.metrics),
        reverse=(self._options.best_mode == 'min'),
    )

  def _cleanup_tmp_directories(self):
    utils.cleanup_tmp_directories(
        self.directory,
        primary_host=self._multiprocessing_options.primary_host,
        active_processes=self._multiprocessing_options.active_processes,
        barrier_sync_key_prefix=self._multiprocessing_options.barrier_sync_key_prefix,
    )

  def _get_old_steps_to_remove(self) -> List[int]:
    """Returns checkpoints that should be deleted."""
    # This isn't a duration but there isn't a general counter that we can use so
    # we abuse a duration metric to count the number of steps examined.
    jax.monitoring.record_event_duration_secs(
        '/jax/checkpoint/write/old_steps_examined_count',
        self._checkpoints.size(),
    )
    preservation_result = self._preservation_policy.should_preserve(
        [info for info in self._checkpoints],
        context=preservation_policy_lib.PreservationContext(),
    )
    return [
        checkpoint.step
        for checkpoint, should_preserve in zip(
            self._checkpoints, preservation_result
        )
        if not should_preserve
        and (
            self._options.should_keep_fn is None
            or not self._options.should_keep_fn(checkpoint.step)
        )
    ]

  def _wait_for_checkpointers(self):
    if is_async_checkpointer(self._checkpointer):
      self._checkpointer.wait_until_finished()  # pytype: disable=attribute-error

  def wait_until_finished(self):
    """Blocks until any incomplete save operations are completed.

    Note that this method will typically be a no-op if all checkpointers are
    synchronous, since old checkpoints are already cleaned up immediately after
    completing `save`, and there is no background thread to wait for.

    If some checkpointers are of type :py:class:`.AsyncCheckpointer`, however,
    this method will wait until each of these checkpointers is finished.
    """
    process_index = multihost.process_index()
    current_thread = threading.current_thread()

    if self._finalize_thread.map(
        lambda t: t is None or (not t.is_alive() and t.exception is None)
    ):
      logging.info(
          '[process=%s][thread=%s][wait_until_finished] No Save Finalize'
          ' thread to wait for. Returning.',
          process_index,
          current_thread.name,
      )
      return

    step = self._finalize_thread.get_not_none().step()
    finalize_thread_name = self._finalize_thread.get_not_none().name
    try:
      logging.info(
          '[process=%s][thread=%s][step=%s][wait_until_finished] Waiting for'
          ' Save Finalize thread (%s) to complete.',
          process_index,
          current_thread.name,
          step,
          finalize_thread_name,
      )
      # Let all threads join and wait for the finalize thread to complete.
      # Don't call join() with a lock otherwise we will end up serializing the
      # access to the finalize thread.
      self._finalize_thread.get_not_none().join()
      logging.info(
          '[process=%s][thread=%s][step=%s][wait_until_finished] Done'
          ' waiting for Save Finalize thread (%s) running at step=%d.',
          process_index,
          current_thread.name,
          step,
          finalize_thread_name,
          step,
      )
    except BaseException:  # pylint:disable=broad-exception-caught
      logging.exception(
          '[process=%s][thread=%s][step=%s][wait_until_finished] Save'
          ' Finalize thread (%s) failed.',
          process_index,
          current_thread.name,
          step,
          finalize_thread_name,
      )
      self._checkpoints.delete_if(lambda info: info.step == step)
      raise

  def is_saving_in_progress(self) -> bool:
    """Returns whether a checkpoint save is in progress."""
    processes_saving = self._save_tracker.get_in_progress_ids()
    logging.vlog(
        1,
        '[process=%s][is_saving_in_progress] Processes saving: %s',
        multihost.process_index(),
        processes_saving,
    )
    return bool(processes_saving)

  def check_for_errors(self):
    """Checks for any outstanding errors in completed asynchronous save operations.

    Delegates to underlying Checkpointer.
    """
    if is_async_checkpointer(self._checkpointer):
      self._checkpointer.check_for_errors()  # pytype: disable=attribute-error

  def _finalize_checkpoint(self, step: int):
    """Executes final actions just before the checkpoint write completes.

    * Logs error if any.
    * Records duration saved due to preemption if any.

    Args:
      step: finalized checkpoint step.
    """
    if utils.is_primary_host(self._multiprocessing_options.primary_host):
      try:
        self.check_for_errors()
      except Exception:  # pylint: disable=broad-except
        logging.exception(
            (
                '[step=%s] Checkpointer failed: one or more items may not be'
                ' finalized. Skipping finalization of step checkpoint.'
            ),
            step,
        )
        return None
      # If at a preemption step, record the time since the previous checkpoint.
      # This represents training time that would otherwise have been wasted.
      # If another checkpoint has not been previously saved, measures the time
      # since program start.
      if self.reached_preemption(step):
        if self._checkpoints.size() > 1:
          previous_time = self._checkpoints[-2].time
        else:
          previous_time = _INIT_TIME
        assert not self._checkpoints.empty()
        duration = self._checkpoints[-1].time - previous_time
        jax.monitoring.record_event_duration_secs(
            '/jax/checkpoint/write/preempt/duration_saved_secs',
            duration.total_seconds(),
        )

  def _finalize(self, step: int, steps_to_remove: List[int]):
    """Finalizes individual items and starts garbage collection."""
    process_index = multihost.process_index()
    current_thread = threading.current_thread()
    try:
      self._non_blocking_metadata_store.wait_until_finished()
      self._wait_for_checkpointers()
      # If an error is encountered while waiting for commit futures to complete,
      # we will not proceed past this point.
      self._finalize_checkpoint(step)
      remove_steps_start_time = time.time()
      self._checkpoint_deleter.delete_steps(steps_to_remove)
      jax.monitoring.record_event_duration_secs(
          '/jax/checkpoint/write/remove_steps_duration_secs',
          time.time() - remove_steps_start_time,
      )
    finally:
      self._save_tracker.complete()
      logging.info(
          '[process=%s][thread=%s][step=%s] CheckpointManager Save Finalize is'
          ' syncing with other hosts...',
          process_index,
          current_thread.name,
          step,
      )
      barrier_sync_fn = self._create_thread_safe_barrier_sync_fn()
      barrier_sync_fn(
          multihost.unique_barrier_key(
              'CheckpointManager:finalize',
              prefix=self._multiprocessing_options.barrier_sync_key_prefix,
              suffix=str(step),
          )
      )
      logging.info(
          '[process=%s][thread=%s][step=%s] CheckpointManager Save Finalize is'
          ' done on all hosts.',
          process_index,
          current_thread.name,
          step,
      )

  def close(self):
    """Waits for outstanding operations to finish and closes internal objects."""
    self.wait_until_finished()
    self._checkpointer.close()
    # Call after checkpointer.close().
    self._non_blocking_metadata_store.close()
    self._blocking_metadata_store.close()
    self._checkpoint_deleter.close()

  def __contextmanager__(
      self,
  ) -> Iterable[Self]:
    try:
      yield self
    finally:
      self.close()



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Public API for CheckpointableHandlers."""

# pylint: disable=g-importing-member, g-multiple-import, unused-import, g-bad-import-order

import orbax.checkpoint.experimental.v1._src.handlers.global_registration
from orbax.checkpoint.experimental.v1._src.handlers.types import (
    CheckpointableHandler,
    StatefulCheckpointable,
)

from orbax.checkpoint.experimental.v1._src.handlers.pytree_handler import (
    PyTreeHandler,
)
from orbax.checkpoint.experimental.v1._src.handlers.proto_handler import (
    ProtoHandler,
)
from orbax.checkpoint.experimental.v1._src.handlers.json_handler import (
    JsonHandler,
)


from orbax.checkpoint.experimental.v1._src.handlers.registration import (
    CheckpointableHandlerRegistry,
    global_registry,
    local_registry,
    register_handler,
)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""ArrayLeafHandler that implements the types.LeafHandler Protocol.

The primary purpose of this handler is to provide serialization and
deserialization for jax.Arrays.
"""

import asyncio
import dataclasses
from typing import Awaitable, Protocol, Sequence, cast

from absl import logging
import jax
import jax.experimental.layout as jax_layout
import jax.numpy as jnp
from orbax.checkpoint._src.arrays import types as arrays_types_v0
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.metadata import sharding as sharding_metadata
from orbax.checkpoint._src.metadata import value as value_metadata
from orbax.checkpoint._src.serialization import type_handlers as type_handlers_v0
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.serialization import types


ArraySerializationParam = types.SerializationParam[jax.Array]
ArrayDeserializationParam = types.DeserializationParam["AbstractArray"]
Shape = arrays_types_v0.Shape

if jax.__version_info__ >= (0, 6, 2):
  Format = jax_layout.Format
else:
  Format = jax_layout.Layout


class AbstractArray(Protocol):
  """Abstract representation of an array.

  This is a protocol for an abstract array that can be used to represent various
  metadata types such as jax.ShapeDtypeStruct and ArrayMetadata.

  #TODO(dnlng): All attributes are made optional to support the case where
  # the ArrayMetadata is passed into the metadata() call to pass only the
  # `write_shape`.  Optional attributes are not needed once write_shape is
  # refactored.


  shape:
    Tuple of integers describing the array shape.
  dtype:
    Dtype of array elements.
  Sharding:
    Sharding to indicate how the array is sharded. This can be jax's Sharding or
    Layout or None.
  """

  shape: Shape | None
  dtype: jax.numpy.dtype | None
  sharding: jax.sharding.Sharding | Format | None  # pytype: disable=unsupported-operands


@dataclasses.dataclass
class ArrayMetadata:
  """Array Metadata for the ArrayLeafHandler.

  shape:
    Tuple of integers describing the array shape.
  sharding_metadata:
    ShardingMetadata to indicate how the array is sharded. ShardingMetadata is
    an orbax representation of `jax.sharding.Sharding` which stores the same
    properties but not require accessing real devices.
  dtype:
    Dtype of array elements.
  storage:
    Optional metadata describing how the array is stored in a checkpoint.
  """

  shape: Shape | None
  dtype: jax.numpy.dtype | None
  sharding_metadata: sharding_metadata.ShardingMetadata | None
  storage_metadata: value_metadata.StorageMetadata | None

  @property
  def sharding(self) -> jax.sharding.Sharding | None:
    """Returns the jax sharding from the sharding_metadata if possible.

    This is needed to comply with the AbstractArray protocol.
    """
    if self.sharding_metadata is None:
      return None
    return self.sharding_metadata.to_jax_sharding()


def _create_v0_array_handler(
    context: context_lib.Context,
) -> type_handlers_v0.ArrayHandler:
  """Creates a V0 array handler from a V1 context."""

  saving_options = context.array_options.saving
  primary_host = context.multiprocessing_options.primary_host
  array_handler = type_handlers_v0.ArrayHandler(
      primary_host=primary_host,
      replica_id=None if primary_host is None else 0,
      use_replica_parallel=saving_options.use_replica_parallel,
      enable_write_sharding_file=saving_options.enable_write_sharding_file,
      array_metadata_store=saving_options.array_metadata_store,
  )


  return array_handler


def _create_v0_saving_paraminfo(
    param: ArraySerializationParam,
    context: context_lib.Context,
    serialization_context: types.SerializationContext,
) -> type_handlers_v0.ParamInfo:
  """Creates a V0 ParamInfo from V1 params andn contexts for saving."""

  saving_options = context.array_options.saving

  return type_handlers_v0.ParamInfo(
      name=param.name,
      path=serialization_context.parent_dir.path / param.name,
      parent_dir=serialization_context.parent_dir.path,
      byte_limiter=serialization_context.byte_limiter,
      is_ocdbt_checkpoint=saving_options.use_ocdbt,
      use_zarr3=saving_options.use_zarr3,
      ocdbt_target_data_file_size=saving_options.ocdbt_target_data_file_size,
      ts_context=serialization_context.ts_context,
      value_typestr=None,  # TODO(dnlng): Add value typestr.
      enable_pinned_host_transfer=saving_options.enable_pinned_host_transfer,
  )


def _create_v0_savearg(
    param: ArraySerializationParam,
    context: context_lib.Context,
) -> type_handlers_v0.SaveArgs:
  """Creates a V0 SaveArgs from V1 params and context for saving."""

  fn = context.pytree_options.saving.create_array_storage_options_fn

  if fn:
    storage_options = fn(param.keypath, param.value)
    savearg = type_handlers_v0.SaveArgs(
        dtype=jnp.dtype(storage_options.dtype)
        if storage_options.dtype
        else None,
        chunk_byte_size=storage_options.chunk_byte_size,
        shard_axes=storage_options.shard_axes,
    )
  else:
    savearg = type_handlers_v0.SaveArgs()

  return savearg


def _create_v0_restore_paraminfo(
    param: (
        types.DeserializationParam[None]
        | types.DeserializationParam[AbstractArray]
    ),
    context: context_lib.Context,
    deserialization_context: types.DeserializationContext,
) -> type_handlers_v0.ParamInfo:
  """Creates a V0 ParamInfo from V1 params and contexts for loading."""

  loading_options = context.array_options.Loading

  if isinstance(param.value, ArrayMetadata):
    # the write_shape is populated for metadata() calls.
    v = cast(ArrayMetadata, param.value)
    write_shape = v.storage_metadata.write_shape
  else:
    write_shape = None

  return type_handlers_v0.ParamInfo(
      name=param.name,
      path=deserialization_context.parent_dir / param.name,
      parent_dir=deserialization_context.parent_dir,
      skip_deserialize=False,
      byte_limiter=deserialization_context.byte_limiter,
      is_ocdbt_checkpoint=deserialization_context.ocdbt_checkpoint,
      ts_context=deserialization_context.ts_context,
      raise_array_data_missing_error=loading_options.raise_array_data_missing_error,
      use_zarr3=deserialization_context.zarr3_checkpoint,
      write_shape=write_shape,
  )


def _create_v0_restorearg(
    param: ArrayDeserializationParam,
    context: context_lib.Context,
) -> type_handlers_v0.ArrayRestoreArgs:
  """Creates a V0 ArrayRestoreArgs from V1 params."""

  if param.value is None:
    return type_handlers_v0.ArrayRestoreArgs(restore_type=jax.Array)
  else:
    v = param.value
    assert isinstance(v, (jax.Array, jax.ShapeDtypeStruct, ArrayMetadata))
    return type_handlers_v0.ArrayRestoreArgs(
        restore_type=jax.Array,
        dtype=v.dtype,
        sharding=v.sharding,
        shape=v.shape,
        strict=not context.array_options.loading.enable_padding_and_truncation,
    )


async def _async_futures(commit_futures: Sequence[future.Future]):
  await asyncio.gather(*[asyncio.to_thread(f.result) for f in commit_futures])


class ArrayLeafHandler(types.LeafHandler[jax.Array, AbstractArray]):
  """ArrayLeafHandler that implements the types.LeafHandler Protocol."""

  def __init__(
      self,
      *,
      context: context_lib.Context | None = None,
  ):
    self._context = context_lib.get_context(context)
    self._handler_impl = _create_v0_array_handler(
        self._context,
    )

    logging.vlog(1, "ArrayLeafHandler created.")

  async def serialize(
      self,
      params: Sequence[ArraySerializationParam],
      serialization_context: types.SerializationContext,
  ) -> Awaitable[None]:
    """Serializes jax.Arrays as a checkpointable to a storage location.

    Args:
      params: a sequence of ArraySerializationParam per leaf.
      serialization_context: SerializationContext for the array leaf handler.

    Returns:
      Sequence of commit futures which can be awaited to complete the save
      operation.
    """
    values = [p.value for p in params]
    paraminfos = [
        _create_v0_saving_paraminfo(p, self._context, serialization_context)
        for p in params
    ]
    saveargs = [_create_v0_savearg(p, self._context) for p in params]

    commit_futures = await self._handler_impl.serialize(
        values, paraminfos, saveargs
    )
    assert commit_futures

    return _async_futures(commit_futures)

  async def deserialize(
      self,
      params: Sequence[ArrayDeserializationParam],
      deserialization_context: types.DeserializationContext,
  ) -> Awaitable[Sequence[jax.Array]]:
    """Returns sequence of jax.Arrays from a stored checkpointable location.

    Args:
      params: sequence of ArrayDeserializationParam per leaf.
      deserialization_context: ArrayDeserializationContext for the leaf handler.

    Returns:
      The deserialized sequence of jax.Arrays as leaves.
    """

    # validate all parameters
    paraminfos = [
        _create_v0_restore_paraminfo(p, self._context, deserialization_context)
        for p in params
    ]
    restoreargs = [_create_v0_restorearg(p, self._context) for p in params]

    return asyncio.create_task(
        self._handler_impl.deserialize(paraminfos, restoreargs)
    )

  async def metadata(
      self,
      params: Sequence[types.DeserializationParam[None | AbstractArray]],
      deserialization_context: types.DeserializationContext,
  ) -> Sequence[AbstractArray]:
    """Returns a squence of ArrayMetadata from a stored checkpointable location.

    Args:
      params: sequence of ArrayDeserializationParam per jax.Array leaf.
      deserialization_context: DeserializationContext for the array leaf
        handler.

    Returns:
      Sequence of ArrayMetadata for each provided ArrayDeserializationParam.
    """
    paraminfos = [
        _create_v0_restore_paraminfo(p, self._context, deserialization_context)
        for p in params
    ]

    async def _convert_to_array_metadata() -> Sequence[ArrayMetadata]:
      v0_metadatas = await self._handler_impl.metadata(paraminfos)

      ret = []
      for meta in v0_metadatas:
        array_metadata = ArrayMetadata(
            shape=meta.shape,
            dtype=meta.dtype,
            sharding_metadata=meta.sharding,
            storage_metadata=meta.storage,
        )
        ret.append(array_metadata)

        logging.vlog(1, "array_metadata: %r", array_metadata)

      return ret

    return await _convert_to_array_metadata()



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A signaling client interface and implementations."""

import asyncio
from typing import Protocol, Sequence
from orbax.checkpoint._src.futures import signaling_client


class SignalingClient(Protocol):
  """Client that supports signaling between threads and processes."""

  async def key_value_set(
      self, key: str, value: str, *, allow_overwrite: bool = False
  ):
    """Sets a key-value pair in the client.

    Args:
      key: The key to set.
      value: The value to set.
      allow_overwrite: Whether to allow overwriting an existing value for the
        given key.
    """
    ...

  async def blocking_key_value_get(self, key: str, timeout_secs: int) -> str:
    """Gets the value for a given key in the client.

    Blocks until the key is set or the timeout is reached.

    Args:
      key: The key to get.
      timeout_secs: The timeout in seconds.
    """
    ...

  async def key_value_try_get(self, key: str) -> str | None:
    """Tries to get the value for a given key in the client without blocking.

    Args:
      key: The key to get.
    """
    ...

  async def key_value_delete(self, key: str):
    """Deletes the key-value.

    If the key is a directory, recursively clean up all key-values under the
    directory.

    Args:
      key: The key to delete.
    """
    ...

  async def wait_at_barrier(
      self,
      key: str,
      *,
      timeout_secs: int,
      process_ids: Sequence[int] | None = None,
  ):
    """Waits at a barrier identified by key.

    Args:
      key: The key to wait at.
      timeout_secs: The timeout in seconds.
      process_ids: The participating process ids.
    """
    ...


class _SignalingClient(SignalingClient):
  """An implementation of SignalingClient that wraps V0 implementation."""

  def __init__(self, client: signaling_client.SignalingClient):
    self._client = client

  async def key_value_set(
      self, key: str, value: str, *, allow_overwrite: bool = False
  ):
    return await asyncio.to_thread(
        self._client.key_value_set, key, value, allow_overwrite=allow_overwrite
    )

  async def blocking_key_value_get(self, key: str, timeout_secs: int) -> str:
    return await asyncio.to_thread(
        self._client.blocking_key_value_get, key, timeout_secs
    )

  async def key_value_try_get(self, key: str) -> str | None:
    return await asyncio.to_thread(self._client.key_value_try_get, key)

  async def key_value_delete(self, key: str):
    return await asyncio.to_thread(self._client.key_value_delete, key)

  async def wait_at_barrier(
      self,
      key: str,
      *,
      timeout_secs: int,
      process_ids: Sequence[int] | None = None,
  ):
    return await asyncio.to_thread(
        self._client.wait_at_barrier,
        key,
        timeout_secs=timeout_secs,
        process_ids=process_ids,
    )


def get_signaling_client() -> SignalingClient:
  return _SignalingClient(signaling_client.get_signaling_client())



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Types for path-related constructs."""

from __future__ import annotations

import typing
from typing import Protocol

from etils import epath

Path = epath.Path
PathLike = epath.PathLike


@typing.runtime_checkable
class PathAwaitingCreation(Protocol):
  """A path that may not exist yet, but will exist after `await_creation`.

  This construct is used to represent a path in the process of being created.
  The underlying path can be accessed logically, but the actual location in
  the filesystem should not be accessed until `await_creation` is called.

  Usage::

    path: PathAwaitingCreation = ...
    # Logical accesses are OK.
    print(path.path)
    # Block until the path is known to exist.
    path = await path.await_creation()
    path.exists()  # True.
  """

  def __truediv__(
      self, other: PathAwaitingCreation | PathLike
  ) -> PathAwaitingCreation:
    ...

  @property
  def path(self) -> Path:
    ...

  async def await_creation(self) -> Path:
    ...



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Test base classes for emergency.CheckpointManager."""

# pylint: disable=protected-access

from typing import Any

from absl import logging
from etils import epath
import jax
from jax.experimental import mesh_utils
import numpy as np
from orbax.checkpoint import test_utils
from orbax.checkpoint._src.multihost import multislice
from orbax.checkpoint.experimental.emergency import checkpoint_manager


PyTree = Any

CheckpointManager = checkpoint_manager.CheckpointManager
CheckpointManagerOptions = checkpoint_manager.CheckpointManagerOptions
LocalCheckpointOptions = checkpoint_manager.LocalCheckpointOptions
PersistentCheckpointOptions = checkpoint_manager.PersistentCheckpointOptions


def swap_slices_in_mesh(
    mesh: jax.sharding.Mesh, *, replica_axis_index: int = 0
) -> jax.sharding.Mesh:
  """Reverses the ordering of devices such that slices swap IDs."""
  devices = []
  for slice_id in range(
      multislice.slice_count(mesh, replica_axis_index=replica_axis_index)
  ):
    devices.append(
        multislice.slice_devices(
            mesh, replica_id=slice_id, replica_axis_index=replica_axis_index
        )
    )
  devices.reverse()
  devices = np.stack(devices, axis=replica_axis_index)
  return jax.sharding.Mesh(devices, mesh.axis_names)


def _replace_abstract_array_sharding_with_mesh(
    sds: jax.ShapeDtypeStruct, mesh: jax.sharding.Mesh
) -> jax.ShapeDtypeStruct:
  assert isinstance(sds.sharding, jax.sharding.NamedSharding)
  sds.sharding = jax.sharding.NamedSharding(mesh, sds.sharding.spec)
  return sds


def fill_unspecified_mesh_axes(
    parallelism_vals, target_product, parallelism_type='ICI'
):
  """Evaluates unspecified DCN/ICI parallelism values."""
  if -1 in parallelism_vals:
    assert parallelism_vals.count(-1) == 1, (
        f'Found unspecified values (-1) for more than one {parallelism_type}   '
        '   parallelism axis. At most one axis can be unspecified.'
    )

    determined_val = target_product / np.prod(parallelism_vals) * -1

    assert determined_val >= 1 and determined_val.is_integer, (
        'Unspecified value unable to be determined with the given     '
        f' {parallelism_type} parallelism values'
    )

    parallelism_vals[parallelism_vals.index(-1)] = int(determined_val)

  assert np.prod(parallelism_vals) == target_product, (
      f'Number of devices per slice {target_product} does not match the product'
      f' of the {parallelism_type} parallelism {np.prod(parallelism_vals)}'
  )

  return parallelism_vals


def create_global_mesh(
    num_slices: int,
    *,
    replica_axis_index: int = 0,
    data_parallelism: int | None = None,
    tensor_parallelism: int | None = None,
    fsdp_parallelism: int | None = None,
) -> jax.sharding.Mesh:
  """Creates a global mesh with the given number of slices.

  Args:
    num_slices: Total number of slices. Used to determine DCN parallelism.
    replica_axis_index: The axis of the replica. Tells emergency checkpointing
      the number of slices. Typically 0, but often 1.
    data_parallelism: For ICI parallelism, the parallelism of the data axis.
    tensor_parallelism: For ICI parallelism, the parallelism of the tensor axis.
    fsdp_parallelism: For ICI parallelism, the parallelism of the fsdp axis.

  Returns:
    A global mesh.
  """
  devices = jax.devices()
  num_devices = len(devices)
  num_devices_per_slice = num_devices // num_slices
  tensor_parallelism = tensor_parallelism or 2
  data_parallelism = data_parallelism or 1
  fsdp_parallelism = fsdp_parallelism or -1

  if num_slices <= 1:
    raise ValueError(f'Must run with at least 2 slices. Found: {num_slices}.')

  mesh_axes = [
      'data',
      'stage',
      'fsdp',
      'fsdp_transpose',
      'sequence',
      'tensor',
      'expert',
      'autoregressive',
  ]
  if replica_axis_index != 0:
    if replica_axis_index == 2 or replica_axis_index == 5:
      raise ValueError(
          '`replica_axis` cannot take the place of `fsdp` or `tensor`.'
      )
    swap_axis_name = mesh_axes[replica_axis_index]
    mesh_axes[replica_axis_index] = 'data'
    mesh_axes[0] = swap_axis_name

  # data_parallelism,
  # pipeline_parallelism,
  # fsdp_parallelism,
  # fsdp_transpose_parallelism,
  # sequence_parallelism,
  # tensor_parallelism,
  # expert_parallelism,
  # autoregressive_parallelism,
  dcn_parallelism = [
      1,
      1,
      1,
      1,
      1,
      1,
      1,
      1,
  ]
  dcn_parallelism[replica_axis_index] = num_slices
  ici_parallelism = [
      1,
      1,
      fsdp_parallelism,
      1,
      1,
      tensor_parallelism,
      1,
      1,
  ]
  ici_parallelism[replica_axis_index] = data_parallelism

  # Find possible unspecified parallelisms
  ici_parallelism = fill_unspecified_mesh_axes(
      ici_parallelism, num_devices_per_slice
  )
  devices_array = mesh_utils.create_hybrid_device_mesh(
      ici_parallelism,
      dcn_parallelism,
      devices,
      allow_split_physical_axes=False,
  )
  logging.info(
      'Creating mesh with axes: %s',
      {axis: dim for axis, dim in zip(mesh_axes, devices_array.shape)},
  )
  return jax.sharding.Mesh(devices_array, mesh_axes)


def _log_array_info(name: str, arr: jax.Array):
  logging.info(
      '%s: shape=%s, dtype=%s, sharding=%s',
      name,
      arr.shape,
      arr.dtype,
      arr.sharding,
  )


def create_train_state(global_mesh: jax.sharding.Mesh) -> PyTree:
  """Create a fake train state for testing."""
  dim_size = 32
  train_state = {
      'a_1d': test_utils.create_sharded_array(
          np.arange(dim_size), global_mesh, jax.sharding.PartitionSpec(None)
      ),
      'b_1d': test_utils.create_sharded_array(
          np.arange(dim_size),
          global_mesh,
          jax.sharding.PartitionSpec('tensor'),
      ),
      'c_2d': test_utils.create_sharded_array(
          np.arange(dim_size**2).reshape((dim_size, dim_size)),
          global_mesh,
          jax.sharding.PartitionSpec(None, 'tensor'),
      ),
      'd_2d': test_utils.create_sharded_array(
          np.arange(dim_size**2).reshape((dim_size, dim_size)),
          global_mesh,
          jax.sharding.PartitionSpec('tensor', None),
      ),
      'e_2d': test_utils.create_sharded_array(
          np.arange(dim_size**2).reshape((dim_size, dim_size)),
          global_mesh,
          jax.sharding.PartitionSpec('tensor', 'fsdp'),
      ),
      'f_2d': test_utils.create_sharded_array(
          np.arange(dim_size**2).reshape((dim_size, dim_size)),
          global_mesh,
          jax.sharding.PartitionSpec('fsdp', 'tensor'),
      ),
      'g_2d': test_utils.create_sharded_array(
          np.arange(dim_size**2).reshape((dim_size, dim_size)),
          global_mesh,
          jax.sharding.PartitionSpec(None, None),
      ),
      'h_3d': test_utils.create_sharded_array(
          np.arange(dim_size**3).reshape((dim_size, dim_size, dim_size)),
          global_mesh,
          jax.sharding.PartitionSpec('tensor', None, 'fsdp'),
      ),
      'i_3d': test_utils.create_sharded_array(
          np.arange(dim_size**3).reshape((dim_size, dim_size, dim_size)),
          global_mesh,
          jax.sharding.PartitionSpec(None, None, 'tensor'),
      ),
      'j_3d': test_utils.create_sharded_array(
          np.arange(dim_size**3).reshape((dim_size, dim_size, dim_size)),
          global_mesh,
          jax.sharding.PartitionSpec(None, None, 'fsdp'),
      ),
      'k_3d': test_utils.create_sharded_array(
          np.arange(dim_size**3).reshape((dim_size, dim_size, dim_size)),
          global_mesh,
          jax.sharding.PartitionSpec(None, None, None),
      ),
      'scalar': test_utils.create_sharded_array(
          123, global_mesh, jax.sharding.PartitionSpec()
      ),
      # Taken from failing MaxText run.
      'custom_array': test_utils.create_sharded_array(
          np.arange(8192 * 64).reshape((8192, 64)),
          global_mesh,
          jax.sharding.PartitionSpec('tensor', None),
      ),
  }
  for k, v in train_state.items():
    _log_array_info(k, v)
  return train_state


def create_checkpoint_manager(
    global_mesh: jax.sharding.Mesh,
    abstract_state: PyTree,
    local_directory: epath.Path,
    persistent_directory: epath.Path,
    *,
    use_async: bool = True,
    replica_axis_index: int = 0,
) -> CheckpointManager:
  """Create CheckpointManager for testing."""
  options = CheckpointManagerOptions(
      local=LocalCheckpointOptions(save_interval_steps=2, max_to_keep=2),
      persistent=PersistentCheckpointOptions(
          save_interval_steps=5, max_to_keep=3
      ),
      enable_async_checkpointing=use_async,
      replica_axis_index=replica_axis_index,
  )
  return CheckpointManager(
      local_directory=local_directory,
      persistent_directory=persistent_directory,
      global_mesh=global_mesh,
      abstract_state=abstract_state,
      options=options,
  )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines policies for when a checkpoint is preserved."""

import dataclasses
import datetime
from typing import Any, Callable, Dict, Protocol, Sequence, Set
import numpy as np
from orbax.checkpoint._src.checkpoint_managers import policy_checkpoint_info


NestedDict = Dict[str, Any]
PyTree = Any
PolicyCheckpointInfo = policy_checkpoint_info.PolicyCheckpointInfo


@dataclasses.dataclass(kw_only=True)
class PreservationContext:
  """Additional properties for making a save decision."""


class PreservationPolicy(Protocol):
  """A policy that defines when checkpoints should be preserved.
  """

  def should_preserve(
      self,
      checkpoints: Sequence[PolicyCheckpointInfo],
      *,
      context: PreservationContext,
  ) -> Sequence[bool]:
    """Indicates which checkpoints should be preserved.."""
    ...


@dataclasses.dataclass
class PreserveAll(PreservationPolicy):
  """Preserves all checkpoints."""

  def should_preserve(
      self,
      checkpoints: Sequence[PolicyCheckpointInfo],
      *,
      context: PreservationContext,
  ) -> Sequence[bool]:
    return [True] * len(checkpoints)


@dataclasses.dataclass
class LatestN(PreservationPolicy):
  """Preserves the last n checkpoints. Preserves all checkpoint if n is None."""

  n: int | None = None

  def should_preserve(
      self,
      checkpoints: Sequence[PolicyCheckpointInfo],
      *,
      context: PreservationContext,
  ) -> Sequence[bool]:
    if self.n is None or len(checkpoints) <= self.n:
      return [True] * len(checkpoints)
    return [False] * (len(checkpoints) - self.n) + [True] * self.n


@dataclasses.dataclass
class EveryNSeconds(PreservationPolicy):
  """Ensures checkpoints are preserved at least after the time interval."""

  interval_secs: int

  def should_preserve(
      self,
      checkpoints: Sequence[PolicyCheckpointInfo],
      *,
      context: PreservationContext,
  ) -> Sequence[bool]:
    if not checkpoints:
      return []
    last_preserved_checkpoint = checkpoints[0]
    result = [True]
    for info in checkpoints[1:]:
      if info.time - last_preserved_checkpoint.time >= datetime.timedelta(
          seconds=self.interval_secs
      ):
        last_preserved_checkpoint = info
        result.append(True)
      else:
        result.append(False)
    return result


@dataclasses.dataclass
class EveryNSteps(PreservationPolicy):
  """Ensures checkpoints are preserved at least every N steps."""

  interval_steps: int

  def should_preserve(
      self,
      checkpoints: Sequence[PolicyCheckpointInfo],
      *,
      context: PreservationContext,
  ) -> Sequence[bool]:
    if self.interval_steps == 0:
      raise ValueError('interval_steps must not be 0.')
    return [ckpt.step % self.interval_steps == 0 for ckpt in checkpoints]


@dataclasses.dataclass
class CustomSteps(PreservationPolicy):
  """Preserves checkpoints at the given steps."""

  steps: dataclasses.InitVar[Sequence[int]]
  _steps_set: Set[int] = dataclasses.field(init=False, repr=False)

  def __post_init__(self, steps_init: Sequence[int]):
    """Initializes the internal set of steps after the object is created."""
    self._steps_set = set(steps_init)

  def should_preserve(
      self,
      checkpoints: Sequence[PolicyCheckpointInfo],
      *,
      context: PreservationContext,
  ) -> Sequence[bool]:
    return [ckpt.step in self._steps_set for ckpt in checkpoints]


@dataclasses.dataclass
class AnyPreservationPolicy(PreservationPolicy):
  """Applies multiple preservation policies and preserves if any policy preserves."""

  policies: Sequence[PreservationPolicy]

  def should_preserve(
      self,
      checkpoints: Sequence[PolicyCheckpointInfo],
      *,
      context: PreservationContext,
  ) -> Sequence[bool]:
    should_preserve_by_policy = np.asarray([
        policy.should_preserve(checkpoints, context=context)
        for policy in self.policies
    ])
    return np.any(should_preserve_by_policy, axis=0).tolist()


@dataclasses.dataclass(kw_only=True)
class BestN(PreservationPolicy):
  """A policy that preserves the best checkpoints based on a best_fn.

  get_metric_fn:
    A function that accepts a nested tree of metrics and returns a scalar value
    representing the value used for ranking checkpoints.
  reverse:
    If False (default), checkpoints are sorted in ascending order, according to
    the best_fn. If True, checkpoints are sorted in descending order. Same as
    the semantics of built-in sorted() function.
  n:
    The number of checkpoints to preserve. If None, all checkpoints are
    preserved. If 0, no checkpoints are preserved.
  """

  get_metric_fn: Callable[[PyTree], float]
  reverse: bool = False
  n: int | None = None
  keep_checkpoints_without_metrics: bool = True

  def should_preserve(
      self,
      checkpoints: Sequence[PolicyCheckpointInfo],
      *,
      context: PreservationContext,
  ) -> Sequence[bool]:
    if self.n is None or len(checkpoints) <= self.n:
      return [True] * len(checkpoints)
    if self.n == 0:
      return [False] * len(checkpoints)
    indexed_checkpoints_with_metrics = [
        (i, info)
        for (i, info) in [(i, cp) for i, cp in enumerate(checkpoints)]
        if info.metrics is not None
    ]
    indexed_checkpoints_without_metrics = [
        (i, info)
        for (i, info) in [(i, cp) for i, cp in enumerate(checkpoints)]
        if info.metrics is None
    ]
    indexed_checkpoints_with_metrics = sorted(
        indexed_checkpoints_with_metrics,
        key=lambda item: self.get_metric_fn(item[1].metrics),
        reverse=self.reverse,
    )
    preserve_indices = [
        i for i, _ in indexed_checkpoints_with_metrics[-self.n :]
    ]
    if self.keep_checkpoints_without_metrics:
      preserve_indices += [i for i, _ in indexed_checkpoints_without_metrics]
    preserve_indices_set = set(preserve_indices)
    preserve_flags = [
        i in preserve_indices_set for i in range(len(checkpoints))
    ]
    return preserve_flags



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Synchronization utilities."""

import threading
from typing import Sequence
from absl import logging
from orbax.checkpoint._src.futures import synchronization
from orbax.checkpoint.experimental.v1._src.synchronization import multihost
from orbax.checkpoint.experimental.v1._src.synchronization import signaling_client

HandlerAwaitableSignal = synchronization.HandlerAwaitableSignal
OperationIdGenerator = synchronization.OperationIdGenerator


def _get_unique_barrier_key(
    signal: HandlerAwaitableSignal, operation_id: str
) -> str:
  """Returns a unique barrier key for the signal.

  Args:
    signal: The signal to generate a barrier key for.
    operation_id: The operation id to use as a suffix for the barrier key.

  Returns:
    A unique barrier key for the signal with operation id as key directory.
  """
  return f'{operation_id}/{signal.value}'


async def _get_awaitable_signals_from_contract(
    client: signaling_client.SignalingClient,
    operation_id: str,
) -> Sequence[HandlerAwaitableSignal]:
  """Gets the awaitable signals for the current operation id."""
  barrier_key = _get_unique_barrier_key(
      HandlerAwaitableSignal.AWAITABLE_SIGNALS_CONTRACT,
      operation_id,
  )
  values_str = await client.key_value_try_get(barrier_key)
  if values_str is None:
    # If the key is not found, then there are no awaitable signals.
    return []
  return [HandlerAwaitableSignal(value) for value in values_str.split(',')]


async def await_contracted_signals(operation_id: str):
  """Waits for the contracted signals to be set.

  This function may be called from a background thread, but it assumes that all
  signals that may be awaited must have previously been added to the contract.
  In other words, signals must be added to the contract synchronously, before
  calling this function.

  Args:
    operation_id: The operation id to use for the barrier key. This should be
      obtained on the main thread, synchronously.
  """
  client = signaling_client.get_signaling_client()

  timeout_secs = 600
  receive_signals = await _get_awaitable_signals_from_contract(
      client, operation_id
  )
  for signal in receive_signals:
    logging.vlog(
        1,
        '[process=%d][thread=%s] Waiting for <%s> timeout: %d secs to be set',
        multihost.process_index(),
        threading.current_thread().name,
        signal.value,
        timeout_secs,
    )
    barrier_key = _get_unique_barrier_key(signal, operation_id)
    await client.blocking_key_value_get(barrier_key, timeout_secs * 1000)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for checkpoint format detection."""

from typing import List
from etils import epath
from orbax.checkpoint._src.metadata import checkpoint as checkpoint_metadata

_OCDBT_MANIFEST_FILE = 'manifest.ocdbt'
PYTREE_METADATA_FILE = '_METADATA'
_CHECKPOINT_FILE = 'checkpoint'  # Old msgpack file.


def is_ocdbt_checkpoint(path: epath.PathLike) -> bool:
  """Determines whether a checkpoint uses OCDBT format."""
  path = epath.Path(path)
  return (path / _OCDBT_MANIFEST_FILE).exists()


def _has_msgpack_metadata_file(path: epath.Path) -> bool:
  return (path / _CHECKPOINT_FILE).exists()


def _has_pytree_metadata_file(path: epath.Path) -> bool:
  """Returns True if path contains `_METADATA` or `checkpoint` file."""
  return (path / PYTREE_METADATA_FILE).exists() or _has_msgpack_metadata_file(
      path
  )


def _has_zarray_files(path: epath.Path) -> List[bool]:
  return [(p / '.zarray').exists() for p in path.iterdir() if p.is_dir()]


def _has_tensorstore_data_files(path: epath.Path) -> bool:
  return is_ocdbt_checkpoint(path) or any(_has_zarray_files(path))


def is_pytree_checkpoint(path: epath.Path) -> bool:
  return (
      _has_pytree_metadata_file(path) and _has_tensorstore_data_files(path)
  ) or (
      _has_msgpack_metadata_file(path)
      and all([not x for x in _has_zarray_files(path)])
  )


def is_orbax_checkpoint(path: epath.PathLike) -> bool:
  """Returns whether the path is LIKELY an Orbax checkpoint.

  It is always possible to assemble files within a directory to look like an
  Orbax checkpoint, even if loading would later reject it as invalid. This
  function provides an easy heuristic as to what constitutes an Orbax
  checkpoint.

  Furthermore, Orbax provides a high degree of customizability, so many
  customized edge cases are possible that would not be detected by this
  function. However, most standard use cases should be covered.

  IMPORTANT: The function is unlikely to return a false positive (True when the
  checkpoint is NOT an Orbax checkpoint), but is more likely to return a
  false negative (False when the checkpoint IS an Orbax checkpoint).

  The function is expected to return True when the directory points to a step
  directory (saved by `CheckpointManager`) or an item directory (saved by
  `PyTreeCheckpointHandler`). It will return False when pointing to a root
  directory (containing multiple step directories).


  Args:
    path: A directory corresponding to an Orbax checkpoint.

  Returns:
    True if the path is LIKELY an Orbax checkpoint.

  Raises:
    FileNotFoundError: If the path does not exist.
    NotADirectoryError: If the path is not a directory.
  """
  path = epath.Path(path)
  if not path.exists():
    raise FileNotFoundError(f'Checkpoint path {path} does not exist.')
  if not path.is_dir():
    raise NotADirectoryError(f'Checkpoint path {path} is not a directory.')
  metadata_store = checkpoint_metadata.metadata_store(enable_write=False)
  # Path points to a single step checkpoint with valid metadata.
  if (
      metadata_store.read(checkpoint_metadata.step_metadata_file_path(path))
      is not None
  ):
    return True

  # Path points to valid PyTree checkpoint without newer metadata.
  if is_pytree_checkpoint(path):
    return True

  # Path points to a root directory containing a valid PyTree checkpoint.
  if any([is_pytree_checkpoint(p) for p in path.iterdir()]):
    return True

  return False
